# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import operator
from collections import OrderedDict
from functools import reduce

import torch

import pyro
import pyro.poutine as poutine
from pyro.distributions.util import scale_and_mask
from pyro.infer.elbo import ELBO
from pyro.infer.util import is_validation_enabled, validation_enabled
from pyro.poutine.util import prune_subsample_sites
from pyro.util import check_model_guide_match, check_site_shape, warn_if_nan


def _squared_error(x, y, scale, mask):
    diff = x - y
    if getattr(scale, "shape", ()) or getattr(mask, "shape", ()):
        error = torch.einsum("nbe,nbe->nb", diff, diff)
        return scale_and_mask(error, scale, mask).sum(-1)
    else:
        error = torch.einsum("nbe,nbe->n", diff, diff)
        return scale_and_mask(error, scale, mask)


class EnergyDistance:
    r"""
    Posterior predictive energy distance [1,2] with optional Bayesian
    regularization by the prior.

    Let `p(x,z)=p(z) p(x|z)` be the model, `q(z|x)` be the guide.  Then given
    data `x` and drawing an iid pair of samples :math:`(Z,X)` and
    :math:`(Z',X')` (where `Z` is latent and `X` is the posterior predictive),

    .. math ::

        & Z \sim q(z|x); \quad X \sim p(x|Z) \\
        & Z' \sim q(z|x); \quad X' \sim p(x|Z') \\
        & loss = \mathbb E_X \|X-x\|^\beta
               - \frac 1 2 \mathbb E_{X,X'}\|X-X'\|^\beta
               - \lambda \mathbb E_Z \log p(Z)

    This is a likelihood-free inference algorithm, and can be used for
    likelihoods without tractable density functions. The :math:`\beta` energy
    distance is a robust loss functions, and is well defined for any
    distribution with finite fractional moment :math:`\mathbb E[\|X\|^\beta]`.

    This requires static model structure, a fully reparametrized guide, and
    reparametrized likelihood distributions in the model. Model latent
    distributions may be non-reparametrized.

    **References**

    [1] Gabor J. Szekely, Maria L. Rizzo (2003)
        Energy Statistics: A Class of Statistics Based on Distances.
    [2] Tilmann Gneiting, Adrian E. Raftery (2007)
        Strictly Proper Scoring Rules, Prediction, and Estimation.
        https://www.stat.washington.edu/raftery/Research/PDF/Gneiting2007jasa.pdf

    :param float beta: Exponent :math:`\beta` from [1,2]. The loss function is
        strictly proper for distributions with finite :math:`beta`-absolute moment
        :math:`E[\|X\|^\beta]`. Thus for heavy tailed distributions ``beta`` should
        be small, e.g. for ``Cauchy`` distributions, :math:`\beta<1` is strictly
        proper. Defaults to 1. Must be in the open interval (0,2).
    :param float prior_scale: Nonnegative scale for prior regularization.
        Model parameters are trained only if this is positive.
        If zero (default), then model log densities will not be computed
        (guide log densities are never computed).
    :param int num_particles: The number of particles/samples used to form the
        gradient estimators. Must be at least 2.
    :param int max_plate_nesting: Optional bound on max number of nested
        :func:`pyro.plate` contexts. If omitted, this will guess a valid value
        by running the (model,guide) pair once.
    """

    def __init__(
        self, beta=1.0, prior_scale=0.0, num_particles=2, max_plate_nesting=float("inf")
    ):
        if not (isinstance(beta, (float, int)) and 0 < beta and beta < 2):
            raise ValueError("Expected beta in (0,2), actual {}".format(beta))
        if not (isinstance(prior_scale, (float, int)) and prior_scale >= 0):
            raise ValueError("Expected prior_scale >= 0, actual {}".format(prior_scale))
        if not (isinstance(num_particles, int) and num_particles >= 2):
            raise ValueError(
                "Expected num_particles >= 2, actual {}".format(num_particles)
            )
        self.beta = beta
        self.prior_scale = prior_scale
        self.num_particles = num_particles
        self.vectorize_particles = True
        self.max_plate_nesting = max_plate_nesting

    def _pow(self, x):
        if self.beta == 1:
            return x.sqrt()  # cheaper than .pow()
        return x.pow(self.beta / 2)

    def _get_traces(self, model, guide, args, kwargs):
        if self.max_plate_nesting == float("inf"):
            with validation_enabled(False):  # Avoid calling .log_prob() when undefined.
                # TODO factor this out as a stand-alone helper.
                ELBO._guess_max_plate_nesting(self, model, guide, args, kwargs)
        vectorize = pyro.plate(
            "num_particles_vectorized", self.num_particles, dim=-self.max_plate_nesting
        )

        # Trace the guide as in ELBO.
        with poutine.trace() as tr, vectorize:
            guide(*args, **kwargs)
        guide_trace = tr.trace

        # Trace the model, drawing posterior predictive samples.
        with poutine.trace() as tr, poutine.uncondition():
            with poutine.replay(trace=guide_trace), vectorize:
                model(*args, **kwargs)
        model_trace = tr.trace
        for site in model_trace.nodes.values():
            if site["type"] == "sample" and site["infer"].get("was_observed", False):
                site["is_observed"] = True
        if is_validation_enabled():
            check_model_guide_match(model_trace, guide_trace, self.max_plate_nesting)

        guide_trace = prune_subsample_sites(guide_trace)
        model_trace = prune_subsample_sites(model_trace)
        if is_validation_enabled():
            for site in guide_trace.nodes.values():
                if site["type"] == "sample":
                    warn_if_nan(site["value"], site["name"])
                    if not getattr(site["fn"], "has_rsample", False):
                        raise ValueError(
                            "EnergyDistance requires fully reparametrized guides"
                        )
            for trace in model_trace.nodes.values():
                if site["type"] == "sample":
                    if site["is_observed"]:
                        warn_if_nan(site["value"], site["name"])
                        if not getattr(site["fn"], "has_rsample", False):
                            raise ValueError(
                                "EnergyDistance requires reparametrized likelihoods"
                            )

        if self.prior_scale > 0:
            model_trace.compute_log_prob(
                site_filter=lambda name, site: not site["is_observed"]
            )
            if is_validation_enabled():
                for site in model_trace.nodes.values():
                    if site["type"] == "sample":
                        if not site["is_observed"]:
                            check_site_shape(site, self.max_plate_nesting)

        return guide_trace, model_trace

    def __call__(self, model, guide, *args, **kwargs):
        """
        Computes the surrogate loss that can be differentiated with autograd
        to produce gradient estimates for the model and guide parameters.
        """
        guide_trace, model_trace = self._get_traces(model, guide, args, kwargs)

        # Extract observations and posterior predictive samples.
        data = OrderedDict()
        samples = OrderedDict()
        for name, site in model_trace.nodes.items():
            if site["type"] == "sample" and site["is_observed"]:
                data[name] = site["infer"]["obs"]
                samples[name] = site["value"]
        assert list(data.keys()) == list(samples.keys())
        if not data:
            raise ValueError("Found no observations")

        # Compute energy distance from mean average error and generalized entropy.
        squared_error = []  # E[ (X - x)^2 ]
        squared_entropy = []  # E[ (X - X')^2 ]
        prototype = next(iter(data.values()))
        pairs = (
            prototype.new_ones(self.num_particles, self.num_particles)
            .tril(-1)
            .nonzero(as_tuple=False)
        )
        for name, obs in data.items():
            sample = samples[name]
            scale = model_trace.nodes[name]["scale"]
            mask = model_trace.nodes[name]["mask"]

            # Flatten to subshapes of (num_particles, batch_size, event_size).
            event_dim = model_trace.nodes[name]["fn"].event_dim
            batch_shape = obs.shape[: obs.dim() - event_dim]
            event_shape = obs.shape[obs.dim() - event_dim :]
            if getattr(scale, "shape", ()):
                scale = scale.expand(batch_shape).reshape(-1)
            if getattr(mask, "shape", ()):
                mask = mask.expand(batch_shape).reshape(-1)
            obs = obs.reshape(batch_shape.numel(), event_shape.numel())
            sample = sample.reshape(
                self.num_particles, batch_shape.numel(), event_shape.numel()
            )

            squared_error.append(_squared_error(sample, obs, scale, mask))
            squared_entropy.append(
                _squared_error(*sample[pairs].unbind(1), scale, mask)
            )

        squared_error = reduce(operator.add, squared_error)
        squared_entropy = reduce(operator.add, squared_entropy)
        error = self._pow(squared_error).mean()  # E[ ||X-x||^beta ]
        entropy = self._pow(squared_entropy).mean()  # E[ ||X-X'||^beta ]
        energy = error - 0.5 * entropy

        # Compute prior.
        log_prior = 0
        if self.prior_scale > 0:
            for site in model_trace.nodes.values():
                if site["type"] == "sample" and not site["is_observed"]:
                    log_prior = log_prior + site["log_prob_sum"]

        # Compute final loss.
        loss = energy - self.prior_scale * log_prior
        warn_if_nan(loss, "loss")
        return loss

    def loss(self, *args, **kwargs):
        """
        Not implemented. Added for compatibility with unit tests only.
        """
        raise NotImplementedError("EnergyDistance implements only surrogate loss")



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, List, Optional

from pyro.distributions.torch_distribution import TorchDistributionMixin
from pyro.poutine.messenger import Messenger
from pyro.util import ignore_jit_warnings

if TYPE_CHECKING:
    from pyro.poutine.runtime import Message


class BroadcastMessenger(Messenger):
    """
    Automatically broadcasts the batch shape of the stochastic function
    at a sample site when inside a single or nested plate context.
    The existing `batch_shape` must be broadcastable with the size
    of the :class:`~pyro.plate` contexts installed in the
    `cond_indep_stack`.

    Notice how `model_automatic_broadcast` below automates expanding of
    distribution batch shapes. This makes it easy to modularize a
    Pyro model as the sub-components are agnostic of the wrapping
    :class:`~pyro.plate` contexts.

    >>> def model_broadcast_by_hand():
    ...     with IndepMessenger("batch", 100, dim=-2):
    ...         with IndepMessenger("components", 3, dim=-1):
    ...             sample = pyro.sample("sample", dist.Bernoulli(torch.ones(3) * 0.5)
    ...                                                .expand_by(100))
    ...             assert sample.shape == torch.Size((100, 3))
    ...     return sample

    >>> @poutine.broadcast
    ... def model_automatic_broadcast():
    ...     with IndepMessenger("batch", 100, dim=-2):
    ...         with IndepMessenger("components", 3, dim=-1):
    ...             sample = pyro.sample("sample", dist.Bernoulli(torch.tensor(0.5)))
    ...             assert sample.shape == torch.Size((100, 3))
    ...     return sample
    """

    @staticmethod
    @ignore_jit_warnings(["Converting a tensor to a Python boolean"])
    def _pyro_sample(msg: "Message") -> None:
        """
        :param msg: current message at a trace site.
        """
        if (
            msg["done"]
            or msg["type"] != "sample"
            or not isinstance(msg["fn"], TorchDistributionMixin)
        ):
            return

        dist = msg["fn"]
        actual_batch_shape = dist.batch_shape
        target_batch_shape = [
            None if size == 1 else size for size in actual_batch_shape
        ]
        for f in msg["cond_indep_stack"]:
            if f.dim is None or f.size == -1:
                continue
            assert f.dim < 0
            prefix_batch_shape: List[Optional[int]] = [None] * (
                -f.dim - len(target_batch_shape)
            )
            target_batch_shape = prefix_batch_shape + target_batch_shape
            if (
                target_batch_shape[f.dim] is not None
                and target_batch_shape[f.dim] != f.size
            ):
                raise ValueError(
                    "Shape mismatch inside plate('{}') at site {} dim {}, {} vs {}".format(
                        f.name,
                        msg["name"],
                        f.dim,
                        f.size,
                        target_batch_shape[f.dim],
                    )
                )
            target_batch_shape[f.dim] = f.size
        # Starting from the right, if expected size is None at an index,
        # set it to the actual size if it exists, else 1.
        for i in range(-len(target_batch_shape) + 1, 1):
            if target_batch_shape[i] is None:
                target_batch_shape[i] = (
                    actual_batch_shape[i] if len(actual_batch_shape) >= -i else 1
                )
        msg["fn"] = dist.expand(target_batch_shape)
        if msg["fn"].has_rsample != dist.has_rsample:
            msg["fn"].has_rsample = dist.has_rsample  # copy custom attribute



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Dict, Optional, Tuple, Union

import torch

import pyro.distributions as dist
from pyro.poutine.trace_messenger import TraceMessenger
from pyro.poutine.trace_struct import Trace
from pyro.poutine.util import prune_subsample_sites, site_is_subsample

if TYPE_CHECKING:
    from pyro.distributions.torch_distribution import TorchDistributionMixin
    from pyro.poutine.runtime import Message


class GuideMessenger(TraceMessenger, ABC):
    """
    Abstract base class for effect-based guides.

    Derived classes must implement the :meth:`get_posterior` method.
    """

    def __init__(self, model: Callable) -> None:
        super().__init__()
        # Do not register model as submodule
        self._model = (model,)

    @property
    def model(self) -> Callable:
        return self._model[0]

    def __getstate__(self) -> Dict[str, object]:
        # Avoid pickling the trace.
        state = self.__dict__.copy()
        del state["trace"]
        return state

    def __call__(self, *args, **kwargs) -> Dict[str, torch.Tensor]:  # type: ignore[override]
        """
        Draws posterior samples from the guide and replays the model against
        those samples.

        :returns: A dict mapping sample site name to sample value.
            This includes latent, deterministic, and observed values.
        :rtype: dict
        """
        self.args_kwargs = args, kwargs
        try:
            with self:
                self.model(*args, **kwargs)
        finally:
            del self.args_kwargs

        model_trace, guide_trace = self.get_traces()
        samples = {}
        for name, site in model_trace.nodes.items():
            if site["type"] == "sample":
                assert isinstance(site["value"], torch.Tensor)
                samples[name] = site["value"]
        return samples

    def _pyro_sample(self, msg: "Message") -> None:
        if msg["is_observed"] or site_is_subsample(msg):
            return
        if TYPE_CHECKING:
            assert isinstance(msg["name"], str)
            assert isinstance(msg["fn"], TorchDistributionMixin)
            assert msg["infer"] is not None
        prior = msg["fn"]
        msg["infer"]["prior"] = prior
        posterior = self.get_posterior(msg["name"], prior)
        if isinstance(posterior, torch.Tensor):
            posterior = dist.Delta(posterior, event_dim=prior.event_dim)
        if posterior.batch_shape != prior.batch_shape:
            posterior = posterior.expand(prior.batch_shape)
        msg["fn"] = posterior

    def _pyro_post_sample(self, msg: "Message") -> None:
        # Manually apply outer plates.
        assert msg["infer"] is not None
        prior = msg["infer"].get("prior")
        if prior is not None:
            if TYPE_CHECKING:
                assert isinstance(msg["fn"], TorchDistributionMixin)
            if prior.batch_shape != msg["fn"].batch_shape:
                msg["infer"]["prior"] = prior.expand(msg["fn"].batch_shape)
        return super()._pyro_post_sample(msg)

    @abstractmethod
    def get_posterior(
        self, name: str, prior: "TorchDistributionMixin"
    ) -> Union["TorchDistributionMixin", torch.Tensor]:
        """
        Abstract method to compute a posterior distribution or sample a
        posterior value given a prior distribution conditioned on upstream
        posterior samples.

        Implementations may use ``pyro.param`` and ``pyro.sample`` inside this
        function, but ``pyro.sample`` statements should set
        ``infer={"is_auxiliary": True"}`` .

        Implementations may access further information for computations:

        - ``value = self.upstream_value(name)`` is the value of an upstream
           sample or deterministic site.
        -  ``self.trace`` is a trace of upstream sites, and may be useful for
           other information such as ``self.trace.nodes["my_site"]["fn"]`` or
           ``self.trace.nodes["my_site"]["cond_indep_stack"]`` .
        -  ``args, kwargs = self.args_kwargs`` are the inputs to the model, and
            may be useful for amortization.

        :param str name: The name of the sample site to sample.
        :param prior: The prior distribution of this sample site
            (conditioned on upstream samples from the posterior).
        :type prior: ~pyro.distributions.Distribution
        :returns: A posterior distribution or sample from the posterior
            distribution.
        :rtype: ~pyro.distributions.Distribution or torch.Tensor
        """
        raise NotImplementedError

    def upstream_value(self, name: str) -> Optional[torch.Tensor]:
        """
        For use in :meth:`get_posterior` .

        :returns: The value of an upstream sample or deterministic site
        :rtype: torch.Tensor
        """
        return self.trace.nodes[name]["value"]

    def get_traces(self) -> Tuple[Trace, Trace]:
        """
        This can be called after running :meth:`__call__` to extract a pair of
        traces.

        In contrast to the trace-replay pattern of generating a pair of traces,
        :class:`GuideMessenger` interleaves model and guide computations, so
        only a single ``guide(*args, **kwargs)`` call is needed to create both
        traces. This function merely extract the relevant information from this
        guide's ``.trace`` attribute.

        :returns: a pair ``(model_trace, guide_trace)``
        :rtype: tuple
        """
        guide_trace = prune_subsample_sites(self.trace)
        model_trace = model_trace = guide_trace.copy()
        for name, guide_site in list(guide_trace.nodes.items()):
            if guide_site["type"] != "sample" or guide_site["is_observed"]:
                del guide_trace.nodes[name]
                continue
            model_site = model_trace.nodes[name].copy()
            assert guide_site["infer"] is not None
            model_site["fn"] = guide_site["infer"]["prior"]
            model_trace.nodes[name] = model_site
        return model_trace, guide_trace



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0


from functools import reduce, singledispatch
from typing import TYPE_CHECKING, Any, FrozenSet, Tuple

from typing_extensions import Self

import pyro
from pyro.distributions.distribution import COERCIONS
from pyro.ops.linalg import ignore_torch_deprecation_warnings
from pyro.poutine.runtime import _PYRO_STACK
from pyro.poutine.trace_messenger import TraceMessenger
from pyro.poutine.util import site_is_subsample

# TODO Remove import guard once funsor is a required dependency.
try:
    import funsor
    from funsor.cnf import Contraction
    from funsor.delta import Delta
    from funsor.terms import Funsor, Variable
except ImportError:
    # Create fake types for singledispatch.
    Contraction = type("Contraction", (), {})
    Delta = type("Delta", (), {})
    Funsor = type("Funsor", (), {})
    Variable = type("Variable", (), {})

if TYPE_CHECKING:
    from funsor.distribution import Distribution

    from pyro.poutine.runtime import Message


@singledispatch
def _substitute(x, subs):
    return x


@_substitute.register(str)
def _(x, subs):
    return subs.get(x, x)


@_substitute.register(Variable)
def _(x, subs):
    return subs.get(x.name, x)


@_substitute.register(tuple)
def _(x, subs):
    return tuple(_substitute(part, subs) for part in x)


@singledispatch
def _extract_deltas(f):
    raise NotImplementedError("unmatched {}".format(type(f).__name__))


@_extract_deltas.register(Delta)
def _(f):
    return f


@_extract_deltas.register(Contraction)
def _(f):
    for d in f.terms:
        if isinstance(d, Delta):
            return d


class CollapseMessenger(TraceMessenger):
    """
    EXPERIMENTAL Collapses all sites in the context by lazily sampling and
    attempting to use conjugacy relations. If no conjugacy is known this will
    fail. Code using the results of sample sites must be written to accept
    Funsors rather than Tensors. This requires ``funsor`` to be installed.

    .. warning:: This is not compatible with automatic guessing of
        ``max_plate_nesting``. If any plates appear within the collapsed
        context, you should manually declare ``max_plate_nesting`` to your
        inference algorithm (e.g. ``Trace_ELBO(max_plate_nesting=1)``).
    """

    _coerce = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if CollapseMessenger._coerce is None:
            import funsor
            from funsor.distribution import CoerceDistributionToFunsor

            funsor.set_backend("torch")
            CollapseMessenger._coerce = CoerceDistributionToFunsor("torch")
        self._block = False
        super().__init__(*args, **kwargs)

    def _process_message(self, msg: "Message") -> None:
        if self._block:
            return
        if site_is_subsample(msg):
            return
        super()._process_message(msg)

    def _pyro_sample(self, msg: "Message") -> None:
        # Eagerly convert fn and value to Funsor.
        dim_to_name = {f.dim: f.name for f in msg["cond_indep_stack"]}
        dim_to_name.update(self.preserved_plates)
        msg["fn"] = funsor.to_funsor(msg["fn"], funsor.Real, dim_to_name)
        if TYPE_CHECKING:
            assert isinstance(msg["fn"], Distribution)
        domain = msg["fn"].inputs["value"]
        if msg["value"] is None:
            msg["value"] = funsor.Variable(msg["name"], domain)
        else:
            msg["value"] = funsor.to_funsor(msg["value"], domain, dim_to_name)

        msg["done"] = True
        msg["stop"] = True

    def _pyro_post_sample(self, msg: "Message") -> None:
        if self._block:
            return
        if site_is_subsample(msg):
            return
        super()._pyro_post_sample(msg)

    def _pyro_barrier(self, msg: "Message") -> None:
        # Get log_prob and record factor.
        name, log_prob, log_joint, sampled_vars = self._get_log_prob()
        self._block = True
        pyro.factor(name, log_prob.data)
        self._block = False

        # Sample
        if sampled_vars:
            samples = log_joint.sample(sampled_vars)
            deltas = _extract_deltas(samples)
            samples = {name: point.data for name, (point, _) in deltas.terms}
        else:
            samples = {}

        # Update value.
        assert len(msg["args"]) == 1
        value = msg["args"][0]
        value = _substitute(value, samples)
        msg["value"] = value

    def __enter__(self) -> Self:
        self.preserved_plates = {
            h.dim: h.name for h in _PYRO_STACK if isinstance(h, pyro.plate)
        }
        COERCIONS.append(self._coerce)
        return super().__enter__()

    def __exit__(self, *args) -> None:
        _coerce = COERCIONS.pop()
        assert _coerce is self._coerce
        super().__exit__(*args)

        if any(site["type"] == "sample" for site in self.trace.nodes.values()):
            name, log_prob, _, _ = self._get_log_prob()
            pyro.factor(name, log_prob.data)

    @ignore_torch_deprecation_warnings()
    def _get_log_prob(self) -> Tuple[str, Funsor, Funsor, FrozenSet[str]]:
        # Convert delayed statements to pyro.factor()
        reduced_vars_list = []
        log_prob_terms = []
        plates: FrozenSet[str] = frozenset()
        for name, site in self.trace.nodes.items():
            if not site["is_observed"]:
                reduced_vars_list.append(name)
            log_prob_terms.append(site["fn"](value=site["value"]))
            plates |= frozenset(
                f.name for f in site["cond_indep_stack"] if f.vectorized
            )
        name = reduced_vars_list[0]
        reduced_vars = frozenset(reduced_vars_list)
        assert log_prob_terms, "nothing to collapse"
        self.trace.nodes.clear()
        reduced_plates = plates - frozenset(self.preserved_plates.values())
        if reduced_plates:
            log_prob = funsor.sum_product.sum_product(
                funsor.ops.logaddexp,
                funsor.ops.add,
                log_prob_terms,
                eliminate=reduced_vars | reduced_plates,
                plates=plates,
            )
            log_joint = NotImplemented
        else:
            log_joint = reduce(funsor.ops.add, log_prob_terms)
            log_prob = log_joint.reduce(funsor.ops.logaddexp, reduced_vars)

        return name, log_prob, log_joint, reduced_vars



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import copy
import ctypes
import functools
import numbers
import weakref
from contextlib import contextmanager

import torch
import torch.distributions as torch_dist
from torch import logsumexp
from torch.distributions.utils import broadcast_all

from pyro.util import ignore_jit_warnings

from .. import settings

_VALIDATION_ENABLED = __debug__
torch_dist.Distribution.set_default_validate_args(__debug__)

settings.register("validate_distributions_pyro", __name__, "_VALIDATION_ENABLED")
settings.register(
    "validate_distributions_torch",
    "torch.distributions.distribution",
    "Distribution._validate_args",
)

log_sum_exp = logsumexp  # DEPRECATED


def copy_docs_from(source_class, full_text=False):
    """
    Decorator to copy class and method docs from source to destin class.
    """

    def decorator(destin_class):
        # This works only in python 3.3+:
        # if not destin_class.__doc__:
        #     destin_class.__doc__ = source_class.__doc__
        for name in dir(destin_class):
            if name.startswith("_"):
                continue
            destin_attr = getattr(destin_class, name)
            destin_attr = getattr(destin_attr, "__func__", destin_attr)
            source_attr = getattr(source_class, name, None)
            source_doc = getattr(source_attr, "__doc__", None)
            if source_doc and not getattr(destin_attr, "__doc__", None):
                if full_text or source_doc.startswith("See "):
                    destin_doc = source_doc
                else:
                    destin_doc = "See :meth:`{}.{}.{}`".format(
                        source_class.__module__, source_class.__name__, name
                    )
                if isinstance(destin_attr, property):
                    # Set docs for object properties.
                    # Since __doc__ is read-only, we need to reset the property
                    # with the updated doc.
                    updated_property = property(
                        destin_attr.fget, destin_attr.fset, destin_attr.fdel, destin_doc
                    )
                    setattr(destin_class, name, updated_property)
                else:
                    destin_attr.__doc__ = destin_doc
        return destin_class

    return decorator


# TODO replace with weakref.WeakMethod?
def weakmethod(fn):
    """
    Decorator to enforce weak binding of a method, so as to avoid reference
    cycles when passing a bound method as an argument to other functions.

    In the following example, functional behavior is the same with and without
    the ``@weakmethod`` decorator, but decoration avoids a reference cycle::

        class Foo:
            def __init__(self):
                self.callback = self._callback
            @weakmethod
            def _callback(self, result):
                print(result)
    """

    def weak_fn(weakself, *args, **kwargs):
        self = weakself()
        if self is None:
            raise AttributeError(
                "self was garbage collected when calling self.{}".format(fn.__name__)
            )
        return fn(self, *args, **kwargs)

    @property
    def weak_binder(self):
        weakself = weakref.ref(self)
        return functools.partial(weak_fn, weakself)

    @weak_binder.setter
    def weak_binder(self, new):
        if not (
            isinstance(new, functools.partial)
            and new.func is weak_fn
            and len(new.args) == 1
            and new.args[0] is weakref.ref(self)
        ):
            raise AttributeError("cannot overwrite weakmethod {}".format(fn.__name__))

    return weak_binder


# This helper intervenes in copy.deepcopy's use of a memo dict to
# use .detatch() to copy tensors.
class _DetachMemo(dict):
    def get(self, key, default=None):
        result = super().get(key, default)

        if result is default:
            # Assume key is the id of another object, and look up that object.
            old = ctypes.cast(key, ctypes.py_object).value
            if isinstance(old, torch.Tensor):
                self[key] = result = old.detach()

        return result


def detach(obj):
    """
    Create a deep copy of an object, detaching all :class:`torch.Tensor` s in
    the object. No tensor data is actually copied.

    :param obj: Any python object.
    :returns: A deep copy of ``obj`` with all :class:`torch.Tensor` s detached.
    """
    memo = _DetachMemo()
    return copy.deepcopy(obj, memo)


class _DeepToMemo(dict):
    def __init__(self, to_args, to_kwargs):
        super().__init__()
        self._to_args = to_args
        self._to_kwargs = to_kwargs

    def get(self, key, default=None):
        result = super().get(key, default)
        if result is default:
            # Assume key is the id of another object, and look up that object.
            old = ctypes.cast(key, ctypes.py_object).value
            if isinstance(old, torch.Tensor):
                self[key] = result = old.to(*self._to_args, **self._to_kwargs)
            elif isinstance(old, torch.nn.Module):
                old.to(*self._to_args, **self._to_kwargs)
        return result


def deep_to(obj, *args, **kwargs):
    r"""
    Create a deep copy of an arbitrary Python object, calling ``.to(*args,
    **kwargs)`` on all :class:`torch.Tensor` s and :class:`torch.nn.Module` s
    in that object.

    Like :meth:`torch.Tensor.to` but unlike :meth:`torch.nn.Module.to`, this
    creates new objects. For compatibility with existing PyTorch module
    behavior, this first calls :meth:`torch.nn.Module.to` in-place, then
    creates a deep copy of the module.

    :param obj: Any python object.
    :param \*args:
    :param \*\*kwargs: See :meth:`torch.Tensor.to`.
    :returns: A deep copy of ``obj`` with all :class:`torch.Tensor` s and
        :class:`torch.nn.Module` s mapped.
    """
    memo = _DeepToMemo(args, kwargs)
    return copy.deepcopy(obj, memo)


# Import a lazy jit.script decorator only if available.
torch_jit_script_if_tracing = getattr(
    torch.jit,
    "script_if_tracing",
    getattr(torch.jit, "_script_if_tracing", torch.jit.script),
)


def is_identically_zero(x):
    """
    Check if argument is exactly the number zero. True for the number zero;
    false for other numbers; false for :class:`~torch.Tensor`s.
    """
    if isinstance(x, numbers.Number):
        return x == 0
    if not torch._C._get_tracing_state():
        if isinstance(x, torch.Tensor) and x.dtype == torch.int64 and not x.shape:
            return x.item() == 0
    return False


def is_identically_one(x):
    """
    Check if argument is exactly the number one. True for the number one;
    false for other numbers; false for :class:`~torch.Tensor`s.
    """
    if isinstance(x, numbers.Number):
        return x == 1
    if not torch._C._get_tracing_state():
        if isinstance(x, torch.Tensor) and x.dtype == torch.int64 and not x.shape:
            return x.item() == 1
    return False


def broadcast_shape(*shapes, **kwargs):
    """
    Similar to ``np.broadcast()`` but for shapes.
    Equivalent to ``np.broadcast(*map(np.empty, shapes)).shape``.

    :param tuple shapes: shapes of tensors.
    :param bool strict: whether to use extend-but-not-resize broadcasting.
    :returns: broadcasted shape
    :rtype: tuple
    :raises: ValueError
    """
    strict = kwargs.pop("strict", False)
    reversed_shape = []
    for shape in shapes:
        for i, size in enumerate(reversed(shape)):
            if i >= len(reversed_shape):
                reversed_shape.append(size)
            elif reversed_shape[i] == 1 and not strict:
                reversed_shape[i] = size
            elif reversed_shape[i] != size and (size != 1 or strict):
                raise ValueError(
                    "shape mismatch: objects cannot be broadcast to a single shape: {}".format(
                        " vs ".join(map(str, shapes))
                    )
                )
    return tuple(reversed(reversed_shape))


def gather(value, index, dim):
    """
    Broadcasted gather of indexed values along a named dim.
    """
    value, index = broadcast_all(value, index)
    with ignore_jit_warnings():
        zero = torch.zeros(1, dtype=torch.long, device=index.device)
    index = index.index_select(dim, zero)
    return value.gather(dim, index)


def sum_rightmost(value, dim):
    """
    Sum out ``dim`` many rightmost dimensions of a given tensor.

    If ``dim`` is 0, no dimensions are summed out.
    If ``dim`` is ``float('inf')``, then all dimensions are summed out.
    If ``dim`` is 1, the rightmost 1 dimension is summed out.
    If ``dim`` is 2, the rightmost two dimensions are summed out.
    If ``dim`` is -1, all but the leftmost 1 dimension is summed out.
    If ``dim`` is -2, all but the leftmost 2 dimensions are summed out.
    etc.

    :param torch.Tensor value: A tensor of ``.dim()`` at least ``dim``.
    :param int dim: The number of rightmost dims to sum out.
    """
    if isinstance(value, numbers.Number):
        return value
    if dim < 0:
        dim += value.dim()
    if dim == 0:
        return value
    if dim >= value.dim():
        return value.sum()
    return value.reshape(value.shape[:-dim] + (-1,)).sum(-1)


def sum_leftmost(value, dim):
    """
    Sum out ``dim`` many leftmost dimensions of a given tensor.

    If ``dim`` is 0, no dimensions are summed out.
    If ``dim`` is ``float('inf')``, then all dimensions are summed out.
    If ``dim`` is 1, the leftmost 1 dimension is summed out.
    If ``dim`` is 2, the leftmost two dimensions are summed out.
    If ``dim`` is -1, all but the rightmost 1 dimension is summed out.
    If ``dim`` is -2, all but the rightmost 2 dimensions are summed out.
    etc.

    Example::

        x = torch.ones(2, 3, 4)
        assert sum_leftmost(x, 1).shape == (3, 4)
        assert sum_leftmost(x, -1).shape == (4,)

    :param torch.Tensor value: A tensor
    :param int dim: Specifies the number of dims to sum out
    """
    if isinstance(value, numbers.Number):
        return value
    if dim < 0:
        dim += value.dim()
    if dim == 0:
        return value
    if dim >= value.dim():
        return value.sum()
    return value.reshape(-1, *value.shape[dim:]).sum(0)


def scale_and_mask(tensor, scale=1.0, mask=None):
    """
    Scale and mask a tensor, broadcasting and avoiding unnecessary ops.

    :param tensor: an input tensor or zero
    :type tensor: torch.Tensor or the number zero
    :param scale: a positive scale
    :type scale: torch.Tensor or number
    :param mask: an optional masking tensor
    :type mask: torch.BoolTensor, bool, or None
    """
    if is_identically_zero(tensor) or (mask is None and is_identically_one(scale)):
        return tensor
    if mask is None or mask is True:
        return tensor * scale
    if mask is False:
        return torch.zeros_like(tensor)
    return torch.where(mask, tensor * scale, tensor.new_zeros(()))


def scalar_like(prototype, fill_value):
    return torch.tensor(fill_value, dtype=prototype.dtype, device=prototype.device)


# work around lack of jit support for torch.eye(..., out=value)
def eye_like(value, m, n=None):
    if n is None:
        n = m
    eye = torch.zeros(m, n, dtype=value.dtype, device=value.device)
    eye.view(-1)[: min(m, n) * n : n + 1] = 1
    return eye


def enable_validation(is_validate):
    global _VALIDATION_ENABLED
    _VALIDATION_ENABLED = is_validate
    torch_dist.Distribution.set_default_validate_args(is_validate)


def is_validation_enabled():
    return _VALIDATION_ENABLED


@contextmanager
def validation_enabled(is_validate=True):
    distribution_validation_status = is_validation_enabled()
    try:
        enable_validation(is_validate)
        yield
    finally:
        enable_validation(distribution_validation_status)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import torch
from torch.utils.data import Dataset

alphabets = {
    "amino-acid": np.array(
        [
            "R",
            "H",
            "K",
            "D",
            "E",
            "S",
            "T",
            "N",
            "Q",
            "C",
            "G",
            "P",
            "A",
            "V",
            "I",
            "L",
            "M",
            "F",
            "Y",
            "W",
        ]
    ),
    "dna": np.array(["A", "C", "G", "T"]),
}


class BiosequenceDataset(Dataset):
    """
    Load biological sequence data, either from a fasta file or a python list.

    :param source: Either the input fasta file path (str) or the input list
        of sequences (list of str).
    :param str source_type: Type of input, either 'list' or 'fasta'.
    :param str alphabet: Alphabet to use. Alphabets 'amino-acid' and 'dna' are
        preset; any other input will be interpreted as the alphabet itself,
        i.e. you can use 'ACGU' for RNA.
    :param int max_length: Total length of the one-hot representation of the
        sequences, including zero padding. Defaults to the maximum sequence
        length in the dataset.
    :param bool include_stop: Append stop symbol to the end of each sequence
        and add the stop symbol to the alphabet.
    :param torch.device device: Device on which data should be stored in
        memory.
    """

    def __init__(
        self,
        source,
        source_type="list",
        alphabet="amino-acid",
        max_length=None,
        include_stop=False,
        device=None,
    ):
        super().__init__()

        # Determine device
        if device is None:
            device = torch.tensor(0.0).device
        self.device = device

        # Get sequences.
        self.include_stop = include_stop
        if source_type == "list":
            seqs = [seq + include_stop * "*" for seq in source]
        elif source_type == "fasta":
            seqs = self._load_fasta(source)

        # Get lengths.
        self.L_data = torch.tensor([float(len(seq)) for seq in seqs], device=device)
        if max_length is None:
            self.max_length = int(torch.max(self.L_data))
        else:
            self.max_length = max_length
        self.data_size = len(self.L_data)

        # Get alphabet.
        if alphabet in alphabets:
            alphabet = alphabets[alphabet]
        else:
            alphabet = np.array(list(alphabet))
        if self.include_stop:
            alphabet = np.array(list(alphabet) + ["*"])
        self.alphabet = alphabet
        self.alphabet_length = len(alphabet)

        # Build dataset.
        self.seq_data = torch.cat(
            [self._one_hot(seq, alphabet, self.max_length).unsqueeze(0) for seq in seqs]
        )

    def _load_fasta(self, source):
        """A basic multiline fasta parser."""
        seqs = []
        seq = ""
        with open(source, "r") as fr:
            for line in fr:
                if line[0] == ">":
                    if seq != "":
                        if self.include_stop:
                            seq += "*"
                        seqs.append(seq)
                        seq = ""
                else:
                    seq += line.strip("\n")
        if seq != "":
            if self.include_stop:
                seq += "*"
            seqs.append(seq)
        return seqs

    def _one_hot(self, seq, alphabet, length):
        """One hot encode and pad with zeros to max length."""
        # One hot encode.
        oh = torch.tensor(
            (np.array(list(seq))[:, None] == alphabet[None, :]).astype(np.float64),
            device=self.device,
        )
        # Pad.
        x = torch.cat(
            [oh, torch.zeros([length - len(seq), len(alphabet)], device=self.device)]
        )

        return x

    def __len__(self):
        return self.data_size

    def __getitem__(self, ind):
        return (self.seq_data[ind], self.L_data[ind])


def write(x, alphabet, file, truncate_stop=False, append=False, scores=None):
    """
    Write sequence samples to file.

    :param ~torch.Tensor x: One-hot encoded sequences, with size
        ``(data_size, seq_length, alphabet_length)``. May be padded with
        zeros for variable length sequences.
    :param ~np.array alphabet: Alphabet.
    :param str file: Output file, where sequences will be written
        in fasta format.
    :param bool truncate_stop: If True, sequences will be truncated at the
        first stop symbol (i.e. the stop symbol and everything after will not
        be written). If False, the whole sequence will be written, including
        any internal stop symbols.
    :param bool append: If True, sequences are appended to the end of the
        output file. If False, the file is first erased.
    """
    print_alphabet = np.array(list(alphabet) + [""])
    x = torch.cat([x, torch.zeros(list(x.shape[:2]) + [1])], -1)
    if truncate_stop:
        mask = (
            torch.cumsum(
                torch.matmul(
                    x, torch.tensor(print_alphabet == "*", dtype=torch.double)
                ),
                -1,
            )
            > 0
        ).to(torch.double)
        x = x * (1 - mask).unsqueeze(-1)
        x[:, :, -1] = mask
    else:
        x[:, :, -1] = (torch.sum(x, -1) < 0.5).to(torch.double)
    index = (
        torch.matmul(x, torch.arange(x.shape[-1], dtype=torch.double))
        .to(torch.long)
        .cpu()
        .numpy()
    )
    if scores is None:
        seqs = [
            ">{}\n".format(j) + "".join(elem) + "\n"
            for j, elem in enumerate(print_alphabet[index])
        ]
    else:
        seqs = [
            ">{}\n".format(j) + "".join(elem) + "\n"
            for j, elem in zip(scores, print_alphabet[index])
        ]
    if append:
        open_flag = "a"
    else:
        open_flag = "w"
    with open(file, open_flag) as fw:
        fw.write("".join(seqs))



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
from torch.distributions.utils import lazy_property
from torch.nn.functional import pad

from pyro.distributions.multivariate_studentt import MultivariateStudentT
from pyro.distributions.torch import MultivariateNormal
from pyro.distributions.util import broadcast_shape
from pyro.ops.tensor_utils import precision_to_scale_tril


class Gamma:
    """
    Non-normalized Gamma distribution.

        Gamma(concentration, rate) ~ (concentration - 1) * log(s) - rate * s
    """

    def __init__(self, log_normalizer, concentration, rate):
        self.log_normalizer = log_normalizer
        self.concentration = concentration
        self.rate = rate

    def log_density(self, s):
        """
        Non-normalized log probability of Gamma distribution.

        This is mainly used for testing.
        """
        return self.log_normalizer + (self.concentration - 1) * s.log() - self.rate * s

    def logsumexp(self):
        """
        Integrates out the latent variable.
        """
        return (
            self.log_normalizer
            + torch.lgamma(self.concentration)
            - self.concentration * self.rate.log()
        )


class GammaGaussian:
    """
    Non-normalized GammaGaussian distribution:

        GammaGaussian(x, s) ~ (concentration + 0.5 * dim - 1) * log(s)
                              - rate * s - s * 0.5 * info_vec.T @ inv(precision) @ info_vec)
                              - s * 0.5 * x.T @ precision @ x + s * x.T @ info_vec,

    which will be reparameterized as

        GammaGaussian(x, s) =: alpha * log(s) + s * (-0.5 * x.T @ precision @ x + x.T @ info_vec - beta).

    The `s` variable plays the role of a mixing variable such that

        p(x | s) ~ Gaussian(s * info_vec, s * precision).

    Conditioned on `s`, this represents an arbitrary semidefinite quadratic function,
    which can be interpreted as a rank-deficient Gaussian distribution.
    The precision matrix may have zero eigenvalues, thus it may be impossible
    to work directly with the covariance matrix.

    :param torch.Tensor log_normalizer: a normalization constant, which is mainly used to keep
        track of normalization terms during contractions.
    :param torch.Tensor info_vec: information vector, which is a scaled version of the mean
        ``info_vec = precision @ mean``. We use this represention to make gaussian contraction
        fast and stable.
    :param torch.Tensor precision: precision matrix of this gaussian.
    :param torch.Tensor alpha: reparameterized shape parameter of the marginal Gamma distribution of
        `s`. The shape parameter Gamma.concentration is reparameterized by:

            alpha = Gamma.concentration + 0.5 * dim - 1

    :param torch.Tensor beta: reparameterized rate parameter of the marginal Gamma distribution of
        `s`. The rate parameter Gamma.rate is reparameterized by:

            beta = Gamma.rate + 0.5 * info_vec.T @ inv(precision) @ info_vec
    """

    def __init__(self, log_normalizer, info_vec, precision, alpha, beta):
        # NB: using info_vec instead of mean to deal with rank-deficient problem
        assert info_vec.dim() >= 1
        assert precision.dim() >= 2
        assert precision.shape[-2:] == info_vec.shape[-1:] * 2
        self.log_normalizer = log_normalizer
        self.info_vec = info_vec
        self.precision = precision
        self.alpha = alpha
        self.beta = beta

    def dim(self):
        return self.info_vec.size(-1)

    @lazy_property
    def batch_shape(self):
        return broadcast_shape(
            self.log_normalizer.shape,
            self.info_vec.shape[:-1],
            self.precision.shape[:-2],
            self.alpha.shape,
            self.beta.shape,
        )

    def expand(self, batch_shape):
        n = self.dim()
        log_normalizer = self.log_normalizer.expand(batch_shape)
        info_vec = self.info_vec.expand(batch_shape + (n,))
        precision = self.precision.expand(batch_shape + (n, n))
        alpha = self.alpha.expand(batch_shape)
        beta = self.beta.expand(batch_shape)
        return GammaGaussian(log_normalizer, info_vec, precision, alpha, beta)

    def reshape(self, batch_shape):
        n = self.dim()
        log_normalizer = self.log_normalizer.reshape(batch_shape)
        info_vec = self.info_vec.reshape(batch_shape + (n,))
        precision = self.precision.reshape(batch_shape + (n, n))
        alpha = self.alpha.reshape(batch_shape)
        beta = self.beta.reshape(batch_shape)
        return GammaGaussian(log_normalizer, info_vec, precision, alpha, beta)

    def __getitem__(self, index):
        """
        Index into the batch_shape of a GammaGaussian.
        """
        assert isinstance(index, tuple)
        log_normalizer = self.log_normalizer[index]
        info_vec = self.info_vec[index + (slice(None),)]
        precision = self.precision[index + (slice(None), slice(None))]
        alpha = self.alpha[index]
        beta = self.beta[index]
        return GammaGaussian(log_normalizer, info_vec, precision, alpha, beta)

    @staticmethod
    def cat(parts, dim=0):
        """
        Concatenate a list of GammaGaussians along a given batch dimension.
        """
        if dim < 0:
            dim += len(parts[0].batch_shape)
        args = [
            torch.cat([getattr(g, attr) for g in parts], dim=dim)
            for attr in ["log_normalizer", "info_vec", "precision", "alpha", "beta"]
        ]
        return GammaGaussian(*args)

    def event_pad(self, left=0, right=0):
        """
        Pad along event dimension.
        """
        lr = (left, right)
        info_vec = pad(self.info_vec, lr)
        precision = pad(self.precision, lr + lr)
        # no change for alpha, beta because we are working with reparameterized version;
        # otherwise, we need to change alpha (similar for beta) to
        # keep the term (alpha + 0.5 * dim - 1) * log(s) constant
        # (note that `dim` has been changed due to padding)
        return GammaGaussian(
            self.log_normalizer, info_vec, precision, self.alpha, self.beta
        )

    def event_permute(self, perm):
        """
        Permute along event dimension.
        """
        assert isinstance(perm, torch.Tensor)
        assert perm.shape == (self.dim(),)
        info_vec = self.info_vec[..., perm]
        precision = self.precision[..., perm][..., perm, :]
        return GammaGaussian(
            self.log_normalizer, info_vec, precision, self.alpha, self.beta
        )

    def __add__(self, other):
        """
        Adds two GammaGaussians in log-density space.
        """
        assert isinstance(other, GammaGaussian)
        assert self.dim() == other.dim()
        return GammaGaussian(
            self.log_normalizer + other.log_normalizer,
            self.info_vec + other.info_vec,
            self.precision + other.precision,
            self.alpha + other.alpha,
            self.beta + other.beta,
        )

    def log_density(self, value, s):
        """
        Evaluate the log density of this GammaGaussian at a point value::

            alpha * log(s) + s * (-0.5 * value.T @ precision @ value + value.T @ info_vec - beta) + log_normalizer

        This is mainly used for testing.
        """
        if value.size(-1) == 0:
            batch_shape = broadcast_shape(value.shape[:-1], s.shape, self.batch_shape)
            return (
                self.alpha * s.log()
                - self.beta * s
                + self.log_normalizer.expand(batch_shape)
            )
        result = (-0.5) * self.precision.matmul(value.unsqueeze(-1)).squeeze(-1)
        result = result + self.info_vec
        result = (value * result).sum(-1)
        return self.alpha * s.log() + (result - self.beta) * s + self.log_normalizer

    def condition(self, value):
        """
        Condition the Gaussian component on a trailing subset of its state.
        This should satisfy::

            g.condition(y).dim() == g.dim() - y.size(-1)

        Note that since this is a non-normalized Gaussian, we include the
        density of ``y`` in the result. Thus :meth:`condition` is similar to a
        ``functools.partial`` binding of arguments::

            left = x[..., :n]
            right = x[..., n:]
            g.log_density(x, s) == g.condition(right).log_density(left, s)
        """
        assert isinstance(value, torch.Tensor)
        assert value.size(-1) <= self.info_vec.size(-1)

        n = self.dim() - value.size(-1)
        info_a = self.info_vec[..., :n]
        info_b = self.info_vec[..., n:]
        P_aa = self.precision[..., :n, :n]
        P_ab = self.precision[..., :n, n:]
        P_bb = self.precision[..., n:, n:]
        b = value

        info_vec = info_a - P_ab.matmul(b.unsqueeze(-1)).squeeze(-1)
        precision = P_aa

        log_normalizer = self.log_normalizer
        alpha = self.alpha
        beta = (
            self.beta
            + 0.5 * P_bb.matmul(b.unsqueeze(-1)).squeeze(-1).mul(b).sum(-1)
            - b.mul(info_b).sum(-1)
        )
        return GammaGaussian(log_normalizer, info_vec, precision, alpha, beta)

    def marginalize(self, left=0, right=0):
        """
        Marginalizing out variables on either side of the event dimension::

            g.marginalize(left=n).event_logsumexp() = g.event_logsumexp()
            g.marginalize(right=n).event_logsumexp() = g.event_logsumexp()

        and for data ``x``:

            g.condition(x).event_logsumexp().log_density(s)
              = g.marginalize(left=g.dim() - x.size(-1)).log_density(x, s)
        """
        # NB: the easiest way to think about this process is to consider GammaGaussian
        # as a Gaussian with precision and info_vec scaled by `s`.
        if left == 0 and right == 0:
            return self
        if left > 0 and right > 0:
            raise NotImplementedError
        n = self.dim()
        n_b = left + right
        a = slice(left, n - right)  # preserved
        b = slice(None, left) if left else slice(n - right, None)

        P_aa = self.precision[..., a, a]
        P_ba = self.precision[..., b, a]
        P_bb = self.precision[..., b, b]
        P_b = torch.linalg.cholesky(P_bb)
        P_a = torch.linalg.solve_triangular(P_b, P_ba, upper=False)
        P_at = P_a.transpose(-1, -2)
        precision = P_aa - P_at.matmul(P_a)

        info_a = self.info_vec[..., a]
        info_b = self.info_vec[..., b]
        b_tmp = torch.linalg.solve_triangular(P_b, info_b.unsqueeze(-1), upper=False)
        info_vec = info_a
        if n_b < n:
            info_vec = info_vec - P_at.matmul(b_tmp).squeeze(-1)

        alpha = self.alpha - 0.5 * n_b
        beta = self.beta - 0.5 * b_tmp.squeeze(-1).pow(2).sum(-1)
        log_normalizer = (
            self.log_normalizer
            + 0.5 * n_b * math.log(2 * math.pi)
            - P_b.diagonal(dim1=-2, dim2=-1).log().sum(-1)
        )
        return GammaGaussian(log_normalizer, info_vec, precision, alpha, beta)

    def compound(self):
        """
        Integrates out the latent multiplier `s`. The result will be a
        Student-T distribution.
        """
        concentration = self.alpha - 0.5 * self.dim() + 1
        scale_tril = precision_to_scale_tril(self.precision)
        scale_tril_t_u = (
            scale_tril.transpose(-1, -2).matmul(self.info_vec.unsqueeze(-1)).squeeze(-1)
        )
        u_Pinv_u = scale_tril_t_u.pow(2).sum(-1)
        rate = self.beta - 0.5 * u_Pinv_u

        loc = scale_tril.matmul(scale_tril_t_u.unsqueeze(-1)).squeeze(-1)
        scale_tril = scale_tril * (rate / concentration).sqrt().unsqueeze(-1).unsqueeze(
            -1
        )
        return MultivariateStudentT(2 * concentration, loc, scale_tril)

    def event_logsumexp(self):
        """
        Integrates out all latent state (i.e. operating on event dimensions) of Gaussian component.
        """
        n = self.dim()
        chol_P = torch.linalg.cholesky(self.precision)
        chol_P_u = torch.linalg.solve_triangular(
            chol_P, self.info_vec.unsqueeze(-1), upper=False
        ).squeeze(-1)
        u_P_u = chol_P_u.pow(2).sum(-1)
        # considering GammaGaussian as a Gaussian with precision = s * precision, info_vec = s * info_vec,
        # marginalize x variable, we get
        #   logsumexp(s) = alpha * log(s) - s * beta + 0.5 n * log(2 pi) + \
        #       0.5 s * uPu - 0.5 * log|P| - 0.5 n * log(s)
        # use the original parameterization of Gamma, we get
        #   logsumexp(s) = (concentration - 1) * log(s) - s * rate + 0.5 n * log(2 pi) - 0.5 * |P|
        # Note that `(concentration - 1) * log(s) - s * rate` is unnormalized log_prob of
        #   Gamma(concentration, rate)
        concentration = self.alpha - 0.5 * n + 1
        rate = self.beta - 0.5 * u_P_u
        log_normalizer_tmp = 0.5 * n * math.log(2 * math.pi) - chol_P.diagonal(
            dim1=-2, dim2=-1
        ).log().sum(-1)
        return Gamma(self.log_normalizer + log_normalizer_tmp, concentration, rate)


def gamma_and_mvn_to_gamma_gaussian(gamma, mvn):
    """
    Convert a pair of Gamma and Gaussian distributions to a GammaGaussian.

        p(x | s) ~ Gaussian(s * info_vec, s * precision)
        p(s) ~ Gamma(alpha, beta)
        p(x, s) ~ GammaGaussian(info_vec, precision, alpha, beta)

    :param ~pyro.distributions.Gamma gamma: the mixing distribution
    :param ~pyro.distributions.MultivariateNormal mvn: the conditional distribution
        when mixing is 1.
    :return: A GammaGaussian object.
    :rtype: ~pyro.ops.gaussian_gamma.GammaGaussian
    """
    assert isinstance(gamma, torch.distributions.Gamma)
    assert isinstance(mvn, torch.distributions.MultivariateNormal)
    n = mvn.loc.size(-1)
    precision = mvn.precision_matrix
    info_vec = precision.matmul(mvn.loc.unsqueeze(-1)).squeeze(-1)

    # reparameterized version of concentration, rate in GaussianGamma
    alpha = gamma.concentration + (0.5 * n - 1)
    beta = gamma.rate + 0.5 * (info_vec * mvn.loc).sum(-1)
    gaussian_logsumexp = 0.5 * n * math.log(2 * math.pi) + mvn.scale_tril.diagonal(
        dim1=-2, dim2=-1
    ).log().sum(-1)
    log_normalizer = -Gamma(
        gaussian_logsumexp, gamma.concentration, gamma.rate
    ).logsumexp()
    return GammaGaussian(log_normalizer, info_vec, precision, alpha, beta)


def scale_mvn(mvn, s):
    """
    Transforms a MVN distribution to another MVN distribution according to

        scale(mvn(loc, precision), s) := mvn(loc, s * precision).
    """
    assert isinstance(mvn, torch.distributions.MultivariateNormal)
    assert isinstance(s, torch.Tensor)
    batch_shape = broadcast_shape(s.shape, mvn.batch_shape)
    loc = mvn.loc.expand(batch_shape + (-1,))
    # XXX: we might use mvn._unbroadcasted_scale_tril here
    scale_tril = mvn.scale_tril / s.sqrt().unsqueeze(-1).unsqueeze(-1)
    return MultivariateNormal(loc, scale_tril=scale_tril)


def matrix_and_mvn_to_gamma_gaussian(matrix, mvn):
    """
    Convert a noisy affine function to a GammaGaussian, where the noise precision
    is scaled by an auxiliary variable `s`. The noisy affine function (conditioned
    on `s`) is defined as::

        y = x @ matrix + scale(mvn, s).sample()

    :param ~torch.Tensor matrix: A matrix with rightmost shape ``(x_dim, y_dim)``.
    :param ~pyro.distributions.MultivariateNormal mvn: A multivariate normal distribution.
    :return: A GammaGaussian with broadcasted batch shape and ``.dim() == x_dim + y_dim``.
    :rtype: ~pyro.ops.gaussian_gamma.GammaGaussian
    """
    assert isinstance(mvn, torch.distributions.MultivariateNormal)
    assert isinstance(matrix, torch.Tensor)
    x_dim, y_dim = matrix.shape[-2:]
    assert mvn.event_shape == (y_dim,)
    batch_shape = broadcast_shape(matrix.shape[:-2], mvn.batch_shape)
    matrix = matrix.expand(batch_shape + (x_dim, y_dim))
    mvn = mvn.expand(batch_shape)

    P_yy = mvn.precision_matrix
    neg_P_xy = matrix.matmul(P_yy)
    P_xy = -neg_P_xy
    P_yx = P_xy.transpose(-1, -2)
    P_xx = neg_P_xy.matmul(matrix.transpose(-1, -2))
    precision = torch.cat(
        [torch.cat([P_xx, P_xy], -1), torch.cat([P_yx, P_yy], -1)], -2
    )
    info_y = P_yy.matmul(mvn.loc.unsqueeze(-1)).squeeze(-1)
    info_x = -matrix.matmul(info_y.unsqueeze(-1)).squeeze(-1)
    info_vec = torch.cat([info_x, info_y], -1)
    log_normalizer = -0.5 * y_dim * math.log(2 * math.pi) - mvn.scale_tril.diagonal(
        dim1=-2, dim2=-1
    ).log().sum(-1)
    beta = 0.5 * (info_y * mvn.loc).sum(-1)
    alpha = beta.new_full(beta.shape, 0.5 * y_dim)

    result = GammaGaussian(log_normalizer, info_vec, precision, alpha, beta)
    assert result.batch_shape == batch_shape
    assert result.dim() == x_dim + y_dim
    return result


def gamma_gaussian_tensordot(x, y, dims=0):
    """
    Computes the integral over two GammaGaussians:

        `(x @ y)((a,c),s) = log(integral(exp(x((a,b),s) + y((b,c),s)), b))`,

    where `x` is a gaussian over variables (a,b), `y` is a gaussian over variables
    (b,c), (a,b,c) can each be sets of zero or more variables, and `dims` is the size of b.

    :param x: a GammaGaussian instance
    :param y: a GammaGaussian instance
    :param dims: number of variables to contract
    """
    assert isinstance(x, GammaGaussian)
    assert isinstance(y, GammaGaussian)
    na = x.dim() - dims
    nb = dims
    nc = y.dim() - dims
    assert na >= 0
    assert nb >= 0
    assert nc >= 0

    device = x.info_vec.device
    perm = torch.cat(
        [
            torch.arange(na, device=device),
            torch.arange(x.dim(), x.dim() + nc, device=device),
            torch.arange(na, x.dim(), device=device),
        ]
    )
    return (
        (x.event_pad(right=nc) + y.event_pad(left=na))
        .event_permute(perm)
        .marginalize(right=nb)
    )



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

"""
Data loader logic with two main responsibilities:
(i)  download raw data and process; this logic is initiated upon import
(ii) helper functions for dealing with mini-batches, sequence packing, etc.

Data are taken from

Boulanger-Lewandowski, N., Bengio, Y. and Vincent, P.,
"Modeling Temporal Dependencies in High-Dimensional Sequences: Application to
Polyphonic Music Generation and Transcription"

however, the original source of the data seems to be the Institut fuer Algorithmen
und Kognitive Systeme at Universitaet Karlsruhe.
"""

import os
import pickle
from collections import namedtuple
from urllib.request import urlopen

import torch
import torch.nn as nn
from torch.nn.utils.rnn import pad_sequence

from pyro.contrib.examples.util import get_data_directory

dset = namedtuple("dset", ["name", "url", "filename"])

JSB_CHORALES = dset(
    "jsb_chorales",
    "https://d2hg8soec8ck9v.cloudfront.net/datasets/polyphonic/jsb_chorales.pickle",
    "jsb_chorales.pkl",
)

PIANO_MIDI = dset(
    "piano_midi",
    "https://d2hg8soec8ck9v.cloudfront.net/datasets/polyphonic/piano_midi.pickle",
    "piano_midi.pkl",
)

MUSE_DATA = dset(
    "muse_data",
    "https://d2hg8soec8ck9v.cloudfront.net/datasets/polyphonic/muse_data.pickle",
    "muse_data.pkl",
)

NOTTINGHAM = dset(
    "nottingham",
    "https://d2hg8soec8ck9v.cloudfront.net/datasets/polyphonic/nottingham.pickle",
    "nottingham.pkl",
)


# this function processes the raw data; in particular it unsparsifies it
def process_data(base_path, dataset, min_note=21, note_range=88):
    output = os.path.join(base_path, dataset.filename)
    if os.path.exists(output):
        try:
            with open(output, "rb") as f:
                return pickle.load(f)
        except (ValueError, UnicodeDecodeError):
            # Assume python env has changed.
            # Recreate pickle file in this env's format.
            os.remove(output)

    print("processing raw data - {} ...".format(dataset.name))
    data = pickle.load(urlopen(dataset.url))
    processed_dataset = {}
    for split, data_split in data.items():
        processed_dataset[split] = {}
        n_seqs = len(data_split)
        processed_dataset[split]["sequence_lengths"] = torch.zeros(
            n_seqs, dtype=torch.long
        )
        processed_dataset[split]["sequences"] = []
        for seq in range(n_seqs):
            seq_length = len(data_split[seq])
            processed_dataset[split]["sequence_lengths"][seq] = seq_length
            processed_sequence = torch.zeros((seq_length, note_range))
            for t in range(seq_length):
                note_slice = torch.tensor(list(data_split[seq][t])) - min_note
                slice_length = len(note_slice)
                if slice_length > 0:
                    processed_sequence[t, note_slice] = torch.ones(slice_length)
            processed_dataset[split]["sequences"].append(processed_sequence)
    pickle.dump(processed_dataset, open(output, "wb"), pickle.HIGHEST_PROTOCOL)
    print("dumped processed data to %s" % output)


# this logic will be initiated upon import
base_path = get_data_directory(__file__)
if not os.path.exists(base_path):
    os.mkdir(base_path)


# ingest training/validation/test data from disk
def load_data(dataset):
    # download and process dataset if it does not exist
    process_data(base_path, dataset)
    file_loc = os.path.join(base_path, dataset.filename)
    with open(file_loc, "rb") as f:
        dset = pickle.load(f)
        for k, v in dset.items():
            sequences = v["sequences"]
            dset[k]["sequences"] = pad_sequence(sequences, batch_first=True).type(
                torch.Tensor
            )
            dset[k]["sequence_lengths"] = v["sequence_lengths"].to(
                device=torch.Tensor().device
            )
    return dset


# this function takes a torch mini-batch and reverses each sequence
# (w.r.t. the temporal axis, i.e. axis=1).
def reverse_sequences(mini_batch, seq_lengths):
    reversed_mini_batch = torch.zeros_like(mini_batch)
    for b in range(mini_batch.size(0)):
        T = seq_lengths[b]
        time_slice = torch.arange(T - 1, -1, -1, device=mini_batch.device)
        reversed_sequence = torch.index_select(mini_batch[b, :, :], 0, time_slice)
        reversed_mini_batch[b, 0:T, :] = reversed_sequence
    return reversed_mini_batch


# this function takes the hidden state as output by the PyTorch rnn and
# unpacks it it; it also reverses each sequence temporally
def pad_and_reverse(rnn_output, seq_lengths):
    rnn_output, _ = nn.utils.rnn.pad_packed_sequence(rnn_output, batch_first=True)
    reversed_output = reverse_sequences(rnn_output, seq_lengths)
    return reversed_output


# this function returns a 0/1 mask that can be used to mask out a mini-batch
# composed of sequences of length `seq_lengths`
def get_mini_batch_mask(mini_batch, seq_lengths):
    mask = torch.zeros(mini_batch.shape[0:2])
    for b in range(mini_batch.shape[0]):
        mask[b, 0 : seq_lengths[b]] = torch.ones(seq_lengths[b])
    return mask


# this function prepares a mini-batch for training or evaluation.
# it returns a mini-batch in forward temporal order (`mini_batch`) as
# well as a mini-batch in reverse temporal order (`mini_batch_reversed`).
# it also deals with the fact that packed sequences (which are what what we
# feed to the PyTorch rnn) need to be sorted by sequence length.
def get_mini_batch(mini_batch_indices, sequences, seq_lengths, cuda=False):
    # get the sequence lengths of the mini-batch
    seq_lengths = seq_lengths[mini_batch_indices]
    # sort the sequence lengths
    _, sorted_seq_length_indices = torch.sort(seq_lengths)
    sorted_seq_length_indices = sorted_seq_length_indices.flip(0)
    sorted_seq_lengths = seq_lengths[sorted_seq_length_indices]
    sorted_mini_batch_indices = mini_batch_indices[sorted_seq_length_indices]

    # compute the length of the longest sequence in the mini-batch
    T_max = torch.max(seq_lengths)
    # this is the sorted mini-batch
    mini_batch = sequences[sorted_mini_batch_indices, 0:T_max, :]
    # this is the sorted mini-batch in reverse temporal order
    mini_batch_reversed = reverse_sequences(mini_batch, sorted_seq_lengths)
    # get mask for mini-batch
    mini_batch_mask = get_mini_batch_mask(mini_batch, sorted_seq_lengths)

    # cuda() here because need to cuda() before packing
    if cuda:
        mini_batch = mini_batch.cuda()
        mini_batch_mask = mini_batch_mask.cuda()
        mini_batch_reversed = mini_batch_reversed.cuda()

    # do sequence packing
    mini_batch_reversed = nn.utils.rnn.pack_padded_sequence(
        mini_batch_reversed, sorted_seq_lengths, batch_first=True
    )

    return mini_batch, mini_batch_reversed, mini_batch_mask, sorted_seq_lengths



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import itertools
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, defaultdict
from contextlib import ExitStack
from types import SimpleNamespace
from typing import Callable, Dict, Optional, Set, Tuple, Union

import torch
from torch.distributions import biject_to

import pyro
import pyro.distributions as dist
import pyro.poutine as poutine
from pyro.distributions import constraints
from pyro.infer.inspect import get_dependencies, is_sample_site
from pyro.nn.module import PyroModule, PyroParam
from pyro.ops.linalg import ignore_torch_deprecation_warnings
from pyro.poutine.runtime import am_i_wrapped, get_plates
from pyro.poutine.util import site_is_subsample

from .guides import AutoGuide
from .initialization import InitMessenger, init_to_feasible
from .utils import deep_getattr, deep_setattr, helpful_support_errors


# Helper to dispatch to concrete subclasses of AutoGaussian, e.g.
#   AutoGaussian(model, backend="dense")
# is converted to
#   AutoGaussianDense(model)
# The intent is to avoid proliferation of subclasses and docstrings,
# and provide a single interface AutoGaussian(...).
class AutoGaussianMeta(type(AutoGuide), ABCMeta):
    backends = {}
    default_backend = "dense"

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert cls.__name__.startswith("AutoGaussian")
        key = cls.__name__.replace("AutoGaussian", "").lower()
        cls.backends[key] = cls

    def __call__(cls, *args, **kwargs):
        if cls is AutoGaussian:
            backend = kwargs.pop("backend", cls.default_backend)
            cls = cls.backends[backend]
        return super(AutoGaussianMeta, cls).__call__(*args, **kwargs)


class AutoGaussian(AutoGuide, metaclass=AutoGaussianMeta):
    """
    Gaussian guide with optimal conditional independence structure.

    This is equivalent to a full rank :class:`AutoMultivariateNormal` guide,
    but with a sparse precision matrix determined by dependencies and plates in
    the model [1]. Depending on model structure, this can have asymptotically
    better statistical efficiency than :class:`AutoMultivariateNormal` .

    This guide implements multiple backends for computation. All backends use
    the same statistically optimal parametrization. The default "dense" backend
    has computational complexity similar to :class:`AutoMultivariateNormal` .
    The experimental "funsor" backend can be asymptotically cheaper in terms of
    time and space (using Gaussian tensor variable elimination [2,3]), but
    incurs large constant overhead. The "funsor" backend requires `funsor
    <https://funsor.pyro.ai>`_ which can be installed via ``pip install
    pyro-ppl[funsor]``.

    The guide currently does not depend on the model's ``*args, **kwargs``.

    Example::

        guide = AutoGaussian(model)
        svi = SVI(model, guide, ...)

    Example using experimental funsor backend::

        !pip install pyro-ppl[funsor]
        guide = AutoGaussian(model, backend="funsor")
        svi = SVI(model, guide, ...)

    **References**

    [1] S.Webb, A.Goliński, R.Zinkov, N.Siddharth, T.Rainforth, Y.W.Teh, F.Wood (2018)
        "Faithful inversion of generative models for effective amortized inference"
        https://dl.acm.org/doi/10.5555/3327144.3327229
    [2] F.Obermeyer, E.Bingham, M.Jankowiak, J.Chiu, N.Pradhan, A.M.Rush, N.Goodman
        (2019)
        "Tensor Variable Elimination for Plated Factor Graphs"
        http://proceedings.mlr.press/v97/obermeyer19a/obermeyer19a.pdf
    [3] F. Obermeyer, E. Bingham, M. Jankowiak, D. Phan, J. P. Chen
        (2019)
        "Functional Tensors for Probabilistic Programming"
        https://arxiv.org/abs/1910.10775

    :param callable model: A Pyro model.
    :param callable init_loc_fn: A per-site initialization function.
        See :ref:`autoguide-initialization` section for available functions.
    :param float init_scale: Initial scale for the standard deviation of each
        (unconstrained transformed) latent variable.
    :param str backend: Back end for performing Gaussian tensor variable
        elimination. Defaults to "dense"; other options include "funsor".
    """

    scale_constraint = constraints.softplus_positive

    def __init__(
        self,
        model: Callable,
        *,
        init_loc_fn: Callable = init_to_feasible,
        init_scale: float = 0.1,
        backend: Optional[str] = None,  # used only by metaclass
    ):
        if not isinstance(init_scale, float) or not (init_scale > 0):
            raise ValueError(f"Expected init_scale > 0. but got {init_scale}")
        self._init_scale = init_scale
        self._original_model = (model,)
        model = InitMessenger(init_loc_fn)(model)
        super().__init__(model)

    @staticmethod
    def _prototype_hide_fn(msg):
        # In contrast to the AutoGuide base class, this includes observation
        # sites and excludes deterministic sites.
        return not is_sample_site(msg)

    def _setup_prototype(self, *args, **kwargs) -> None:
        super()._setup_prototype(*args, **kwargs)

        self.locs = PyroModule()
        self.scales = PyroModule()
        self.white_vecs = PyroModule()
        self.prec_sqrts = PyroModule()
        self._factors = OrderedDict()
        self._plates = OrderedDict()
        self._event_numel = OrderedDict()
        self._unconstrained_event_shapes = OrderedDict()

        # Trace model dependencies.
        model = self._original_model[0]
        self._original_model = None
        self.dependencies = poutine.block(get_dependencies)(model, args, kwargs)[
            "prior_dependencies"
        ]

        # Eliminate observations with no upstream latents.
        for d, upstreams in list(self.dependencies.items()):
            if all(self.prototype_trace.nodes[u]["is_observed"] for u in upstreams):
                del self.dependencies[d]
                del self.prototype_trace.nodes[d]

        # Collect factors and plates.
        for d, site in self.prototype_trace.nodes.items():
            # Prune non-essential parts of the trace to save memory.
            pruned_site, site = site, site.copy()
            pruned_site.clear()

            # Collect factors and plates.
            if site["type"] != "sample" or site_is_subsample(site):
                continue
            assert all(f.vectorized for f in site["cond_indep_stack"])
            self._factors[d] = self._compress_site(site)
            plates = frozenset(site["cond_indep_stack"])
            if site["fn"].batch_shape != _plates_to_shape(plates):
                raise ValueError(
                    f"Shape mismatch at site '{d}'. "
                    "Are you missing a pyro.plate() or .to_event()?"
                )
            if site["is_observed"]:
                # Break irrelevant observation plates.
                plates &= frozenset().union(
                    *(self._plates[u] for u in self.dependencies[d] if u != d)
                )
            self._plates[d] = plates

            # Create location-scale parameters, one per latent variable.
            if site["is_observed"]:
                # This may slightly overestimate, e.g. for Multinomial.
                self._event_numel[d] = site["fn"].event_shape.numel()
                # Account for broken irrelevant observation plates.
                for f in set(site["cond_indep_stack"]) - plates:
                    self._event_numel[d] *= f.size
                continue
            with helpful_support_errors(site):
                init_loc = biject_to(site["fn"].support).inv(site["value"]).detach()
            batch_shape = site["fn"].batch_shape
            event_shape = init_loc.shape[len(batch_shape) :]
            self._unconstrained_event_shapes[d] = event_shape
            self._event_numel[d] = event_shape.numel()
            event_dim = len(event_shape)
            deep_setattr(self.locs, d, PyroParam(init_loc, event_dim=event_dim))
            deep_setattr(
                self.scales,
                d,
                PyroParam(
                    torch.full_like(init_loc, self._init_scale),
                    constraint=self.scale_constraint,
                    event_dim=event_dim,
                ),
            )

        # Create parameters for dependencies, one per factor.
        for d, site in self._factors.items():
            u_size = 0
            for u in self.dependencies[d]:
                if not self._factors[u]["is_observed"]:
                    broken_shape = _plates_to_shape(self._plates[u] - self._plates[d])
                    u_size += broken_shape.numel() * self._event_numel[u]
            d_size = self._event_numel[d]
            if site["is_observed"]:
                d_size = min(d_size, u_size)  # just an optimization
            batch_shape = _plates_to_shape(self._plates[d])

            # Create parameters of each Gaussian factor.
            white_vec = init_loc.new_zeros(batch_shape + (d_size,))
            # We initialize with noise to avoid singular gradient.
            prec_sqrt = torch.rand(
                batch_shape + (u_size, d_size),
                dtype=init_loc.dtype,
                device=init_loc.device,
            )
            prec_sqrt.sub_(0.5).mul_(self._init_scale)
            if not site["is_observed"]:
                # Initialize the [d,d] block to the identity matrix.
                prec_sqrt.diagonal(dim1=-2, dim2=-1).fill_(1)
            deep_setattr(self.white_vecs, d, PyroParam(white_vec, event_dim=1))
            deep_setattr(self.prec_sqrts, d, PyroParam(prec_sqrt, event_dim=2))

    @staticmethod
    def _compress_site(site):
        # Save memory by retaining only necessary parts of the site.
        return {
            "name": site["name"],
            "type": site["type"],
            "cond_indep_stack": site["cond_indep_stack"],
            "is_observed": site["is_observed"],
            "fn": SimpleNamespace(
                support=site["fn"].support,
                batch_shape=site["fn"].batch_shape,
                event_dim=site["fn"].event_dim,
            ),
        }

    def forward(self, *args, **kwargs) -> Dict[str, torch.Tensor]:
        if self.prototype_trace is None:
            self._setup_prototype(*args, **kwargs)

        aux_values = self._sample_aux_values(temperature=1.0)
        values, log_densities = self._transform_values(aux_values)

        # Replay via Pyro primitives.
        plates = self._create_plates(*args, **kwargs)
        for name, site in self._factors.items():
            if site["is_observed"]:
                continue
            with ExitStack() as stack:
                for frame in site["cond_indep_stack"]:
                    stack.enter_context(plates[frame.name])
                values[name] = pyro.sample(
                    name,
                    dist.Delta(values[name], log_densities[name], site["fn"].event_dim),
                )
        return values

    def median(self, *args, **kwargs) -> Dict[str, torch.Tensor]:
        """
        Returns the posterior median value of each latent variable.

        :return: A dict mapping sample site name to median tensor.
        :rtype: dict
        """
        with torch.no_grad(), poutine.mask(mask=False):
            aux_values = self._sample_aux_values(temperature=0.0)
            values, _ = self._transform_values(aux_values)
        return values

    def _transform_values(
        self,
        aux_values: Dict[str, torch.Tensor],
    ) -> Tuple[Dict[str, torch.Tensor], Union[float, torch.Tensor]]:
        # Learnably transform auxiliary values to user-facing values.
        values = {}
        log_densities = defaultdict(float)
        compute_density = am_i_wrapped() and poutine.get_mask() is not False
        for name, site in self._factors.items():
            if site["is_observed"]:
                continue
            loc = deep_getattr(self.locs, name)
            scale = deep_getattr(self.scales, name)
            unconstrained = aux_values[name] * scale + loc

            # Transform to constrained space.
            transform = biject_to(site["fn"].support)
            values[name] = transform(unconstrained)
            if compute_density:
                assert transform.codomain.event_dim == site["fn"].event_dim
                log_densities[name] = transform.inv.log_abs_det_jacobian(
                    values[name], unconstrained
                ) - scale.log().reshape(site["fn"].batch_shape + (-1,)).sum(-1)

        return values, log_densities

    @abstractmethod
    def _sample_aux_values(self, *, temperature: float) -> Dict[str, torch.Tensor]:
        raise NotImplementedError


class AutoGaussianDense(AutoGaussian):
    """
    Dense implementation of :class:`AutoGaussian` .

    The following are equivalent::

        guide = AutoGaussian(model, backend="dense")
        guide = AutoGaussianDense(model)
    """

    def _setup_prototype(self, *args, **kwargs):
        super()._setup_prototype(*args, **kwargs)

        # Collect global shapes and per-axis indices.
        self._dense_shapes = {}
        global_indices = {}
        pos = 0
        for d, event_shape in self._unconstrained_event_shapes.items():
            batch_shape = self._factors[d]["fn"].batch_shape
            self._dense_shapes[d] = batch_shape, event_shape
            end = pos + (batch_shape + event_shape).numel()
            global_indices[d] = torch.arange(pos, end).reshape(batch_shape + (-1,))
            pos = end
        self._dense_size = pos

        # Create sparse -> dense precision scatter indices.
        self._dense_scatter = {}
        for d, site in self._factors.items():
            prec_sqrt_shape = deep_getattr(self.prec_sqrts, d).shape
            info_vec_shape = prec_sqrt_shape[:-1]
            precision_shape = prec_sqrt_shape[:-1] + prec_sqrt_shape[-2:-1]
            index1 = torch.zeros(info_vec_shape, dtype=torch.long)
            index2 = torch.zeros(precision_shape, dtype=torch.long)

            # Collect local offsets and create index1 for info_vec blockwise.
            upstreams = [
                u for u in self.dependencies[d] if not self._factors[u]["is_observed"]
            ]
            local_offsets = {}
            pos = 0
            for u in upstreams:
                local_offsets[u] = pos
                broken_plates = self._plates[u] - self._plates[d]
                pos += self._event_numel[u] * _plates_to_shape(broken_plates).numel()
                u_index = global_indices[u]

                # Permute broken plates to the right of preserved plates.
                u_index = _break_plates(u_index, self._plates[u], self._plates[d])

                # Scatter global indices into the [u] block.
                u_start = local_offsets[u]
                u_stop = u_start + u_index.size(-1)
                index1[..., u_start:u_stop] = u_index

            # Create index2 for precision blockwise.
            for u, v in itertools.product(upstreams, upstreams):
                u_index = global_indices[u]
                v_index = global_indices[v]

                # Permute broken plates to the right of preserved plates.
                u_index = _break_plates(u_index, self._plates[u], self._plates[d])
                v_index = _break_plates(v_index, self._plates[v], self._plates[d])

                # Scatter global indices into the [u,v] block.
                u_start = local_offsets[u]
                u_stop = u_start + u_index.size(-1)
                v_start = local_offsets[v]
                v_stop = v_start + v_index.size(-1)
                index2[..., u_start:u_stop, v_start:v_stop] = (
                    self._dense_size * u_index.unsqueeze(-1) + v_index.unsqueeze(-2)
                )

            self._dense_scatter[d] = index1.reshape(-1), index2.reshape(-1)

    def _sample_aux_values(self, *, temperature: float) -> Dict[str, torch.Tensor]:
        mvn = self._dense_get_mvn()
        if temperature == 0:
            # Simply return the mode.
            flat_samples = mvn.mean
        elif temperature == 1:
            # Sample from a dense joint Gaussian over flattened variables.
            flat_samples = pyro.sample(
                f"_{self._pyro_name}_latent", mvn, infer={"is_auxiliary": True}
            )
        else:
            raise NotImplementedError(f"Invalid temperature: {temperature}")
        samples = self._dense_unflatten(flat_samples)
        return samples

    def _dense_unflatten(self, flat_samples: torch.Tensor) -> Dict[str, torch.Tensor]:
        # Convert a single flattened sample to a dict of shaped samples.
        sample_shape = flat_samples.shape[:-1]
        samples = {}
        pos = 0
        for d, (batch_shape, event_shape) in self._dense_shapes.items():
            end = pos + (batch_shape + event_shape).numel()
            flat_sample = flat_samples[..., pos:end]
            pos = end
            # Assumes sample shapes are left of batch shapes.
            samples[d] = flat_sample.reshape(
                torch.broadcast_shapes(sample_shape, batch_shape) + event_shape
            )
        return samples

    def _dense_flatten(self, samples: Dict[str, torch.Tensor]) -> torch.Tensor:
        # Convert a dict of shaped samples single flattened sample.
        flat_samples = []
        for d, (batch_shape, event_shape) in self._dense_shapes.items():
            shape = samples[d].shape
            sample_shape = shape[: len(shape) - len(batch_shape) - len(event_shape)]
            flat_samples.append(samples[d].reshape(sample_shape + (-1,)))
        return torch.cat(flat_samples, dim=-1)

    def _dense_get_mvn(self):
        # Create a dense joint Gaussian over flattened variables.
        flat_info_vec = torch.zeros(self._dense_size)
        flat_precision = torch.zeros(self._dense_size**2)
        for d, (index1, index2) in self._dense_scatter.items():
            white_vec = deep_getattr(self.white_vecs, d)
            prec_sqrt = deep_getattr(self.prec_sqrts, d)
            info_vec = (prec_sqrt @ white_vec[..., None])[..., 0]
            precision = prec_sqrt @ prec_sqrt.transpose(-1, -2)
            flat_info_vec.scatter_add_(0, index1, info_vec.reshape(-1))
            flat_precision.scatter_add_(0, index2, precision.reshape(-1))
        info_vec = flat_info_vec
        precision = flat_precision.reshape(self._dense_size, self._dense_size)
        scale_tril = _precision_to_scale_tril(precision)
        loc = (
            scale_tril @ (scale_tril.transpose(-1, -2) @ info_vec.unsqueeze(-1))
        ).squeeze(-1)
        return dist.MultivariateNormal(loc, scale_tril=scale_tril)


class AutoGaussianFunsor(AutoGaussian):
    """
    Funsor implementation of :class:`AutoGaussian` .

    The following are equivalent::
        guide = AutoGaussian(model, backend="funsor")
        guide = AutoGaussianFunsor(model)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _import_funsor()

    def _setup_prototype(self, *args, **kwargs):
        super()._setup_prototype(*args, **kwargs)
        funsor = _import_funsor()

        # Check TVE condition 1: plate nesting is monotone.
        for d in self._factors:
            pd = {p.name for p in self._plates[d]}
            for u in self.dependencies[d]:
                pu = {p.name for p in self._plates[u]}
                if pu <= pd:
                    continue  # ok
                raise NotImplementedError(
                    "Expected monotone plate nesting, but found dependency "
                    f"{repr(u)} -> {repr(d)} leaves plates {pu - pd}. "
                    "Consider splitting into multiple guides via AutoGuideList, "
                    "or replacing the plate in the model by .to_event()."
                )

        # Determine TVE problem shape.
        factor_inputs: Dict[str, OrderedDict[str, funsor.Domain]] = {}
        eliminate: Set[str] = set()
        plate_to_dim: Dict[str, int] = {}
        for d, site in self._factors.items():
            inputs = OrderedDict()
            for f in sorted(self._plates[d], key=lambda f: f.dim):
                plate_to_dim[f.name] = f.dim
                inputs[f.name] = funsor.Bint[f.size]
                eliminate.add(f.name)
            for u in self.dependencies[d]:
                if self._factors[u]["is_observed"]:
                    continue
                inputs[u] = funsor.Reals[self._unconstrained_event_shapes[u]]
                eliminate.add(u)
            factor_inputs[d] = inputs

        self._funsor_factor_inputs = factor_inputs
        self._funsor_eliminate = frozenset(eliminate)
        self._funsor_plate_to_dim = plate_to_dim
        self._funsor_plates = frozenset(plate_to_dim)

    def _sample_aux_values(self, *, temperature: float) -> Dict[str, torch.Tensor]:
        funsor = _import_funsor()

        # Convert torch to funsor.
        particle_plates = frozenset(get_plates())
        plate_to_dim = self._funsor_plate_to_dim.copy()
        plate_to_dim.update({f.name: f.dim for f in particle_plates})
        factors = {}
        for d, inputs in self._funsor_factor_inputs.items():
            batch_shape = torch.Size(
                p.size for p in sorted(self._plates[d], key=lambda p: p.dim)
            )
            white_vec = deep_getattr(self.white_vecs, d)
            prec_sqrt = deep_getattr(self.prec_sqrts, d)
            factors[d] = funsor.gaussian.Gaussian(
                white_vec=white_vec.reshape(batch_shape + white_vec.shape[-1:]),
                prec_sqrt=prec_sqrt.reshape(batch_shape + prec_sqrt.shape[-2:]),
                inputs=inputs,
            )

        # Perform Gaussian tensor variable elimination.
        if temperature == 1:
            samples, log_prob = _try_possibly_intractable(
                funsor.recipes.forward_filter_backward_rsample,
                factors=factors,
                eliminate=self._funsor_eliminate,
                plates=frozenset(plate_to_dim),
                sample_inputs={f.name: funsor.Bint[f.size] for f in particle_plates},
            )

        else:
            samples, log_prob = _try_possibly_intractable(
                funsor.recipes.forward_filter_backward_precondition,
                factors=factors,
                eliminate=self._funsor_eliminate,
                plates=frozenset(plate_to_dim),
            )

            # Substitute noise.
            sample_shape = torch.Size(f.size for f in particle_plates)
            noise = torch.randn(sample_shape + log_prob.inputs["aux"].shape)
            noise.mul_(temperature)
            aux = funsor.Tensor(noise)[tuple(f.name for f in particle_plates)]
            with funsor.interpretations.memoize():
                samples = {k: v(aux=aux) for k, v in samples.items()}
                log_prob = log_prob(aux=aux)

        # Convert funsor to torch.
        if am_i_wrapped() and poutine.get_mask() is not False:
            log_prob = funsor.to_data(log_prob, name_to_dim=plate_to_dim)
            pyro.factor(f"_{self._pyro_name}_latent", log_prob, has_rsample=True)
        samples = {
            k: funsor.to_data(v, name_to_dim=plate_to_dim) for k, v in samples.items()
        }
        return samples


def _precision_to_scale_tril(P):
    # Ref: https://nbviewer.jupyter.org/gist/fehiepsi/5ef8e09e61604f10607380467eb82006#Precision-to-scale_tril
    Lf = torch.linalg.cholesky(torch.flip(P, (-2, -1)))
    L_inv = torch.transpose(torch.flip(Lf, (-2, -1)), -2, -1)
    L = torch.linalg.solve_triangular(
        L_inv, torch.eye(P.shape[-1], dtype=P.dtype, device=P.device), upper=False
    )
    return L


@ignore_torch_deprecation_warnings()
def _try_possibly_intractable(fn, *args, **kwargs):
    # Convert ValueError into NotImplementedError.
    try:
        return fn(*args, **kwargs)
    except ValueError as e:
        if str(e) != "intractable!":
            raise e from None
        raise NotImplementedError(
            "Funsor backend found intractable plate nesting. "
            'Consider using AutoGaussian(..., backend="dense"), '
            "splitting into multiple guides via AutoGuideList, or "
            "replacing some plates in the model by .to_event()."
        ) from e


def _plates_to_shape(plates):
    shape = [1] * max([0] + [-f.dim for f in plates])
    for f in plates:
        shape[f.dim] = f.size
    return torch.Size(shape)


def _break_plates(x, all_plates, kept_plates):
    """
    Reshapes and permutes a tensor ``x`` with event_dim=1 and batch shape given
    by ``all_plates`` by breaking all plates not in ``kept_plates``. Each
    broken plate is moved into the event shape, and finally the event shape is
    flattend back to a single dimension.
    """
    assert x.shape[:-1] == _plates_to_shape(all_plates)  # event_dim == 1
    kept_plates = kept_plates & all_plates
    broken_plates = all_plates - kept_plates

    if not broken_plates:
        return x

    if not kept_plates:
        # Empty batch shape.
        return x.reshape(-1)

    batch_shape = _plates_to_shape(kept_plates)
    if max(p.dim for p in kept_plates) < min(p.dim for p in broken_plates):
        # No permutation is necessary.
        return x.reshape(batch_shape + (-1,))

    # We need to permute broken plates left past kept plates.
    event_dims = {-1} | {p.dim - 1 for p in broken_plates}
    perm = sorted(range(-x.dim(), 0), key=lambda d: (d in event_dims, d))
    return x.permute(perm).reshape(batch_shape + (-1,))


def _import_funsor():
    try:
        import funsor
    except ImportError as e:
        raise ImportError(
            'AutoGaussian(..., backend="funsor") requires funsor. '
            "Try installing via: pip install pyro-ppl[funsor]"
        ) from e
    funsor.set_backend("torch")
    return funsor


__all__ = [
    "AutoGaussian",
]



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
from torch.distributions import constraints

from pyro.distributions.torch import Exponential
from pyro.distributions.torch_distribution import TorchDistribution


class TruncatedPolyaGamma(TorchDistribution):
    """
    This is a PolyaGamma(1, 0) distribution truncated to have finite support in
    the interval (0, 2.5). See [1] for details. As a consequence of the truncation
    the `log_prob` method is only accurate to about six decimal places. In
    addition the provided sampler is a rough approximation that is only meant to
    be used in contexts where sample accuracy is not important (e.g. in initialization).
    Broadly, this implementation is only intended for usage in cases where good
    approximations of the `log_prob` are sufficient, as is the case e.g. in HMC.

    :param tensor prototype: A prototype tensor of arbitrary shape used to determine
        the `dtype` and `device` returned by `sample` and `log_prob`.

    References

    [1] 'Bayesian inference for logistic models using Polya-Gamma latent variables'
        Nicholas G. Polson, James G. Scott, Jesse Windle.
    """

    truncation_point = 2.5
    num_log_prob_terms = 7
    num_gamma_variates = 8
    assert num_log_prob_terms % 2 == 1

    arg_constraints = {}
    support = constraints.interval(0.0, truncation_point)
    has_rsample = False

    def __init__(self, prototype, validate_args=None):
        self.prototype = prototype
        super(TruncatedPolyaGamma, self).__init__(
            batch_shape=(), event_shape=(), validate_args=validate_args
        )

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(TruncatedPolyaGamma, _instance)
        super(TruncatedPolyaGamma, new).__init__(
            batch_shape, self.event_shape, validate_args=False
        )
        new._validate_args = self.__dict__.get("_validate_args")
        new.prototype = self.prototype
        return new

    def sample(self, sample_shape=()):
        denom = torch.arange(
            0.5, self.num_gamma_variates, device=self.prototype.device
        ).pow(2.0)
        ones = self.prototype.new_ones((self.num_gamma_variates))
        x = Exponential(ones).sample(self.batch_shape + sample_shape)
        x = (x / denom).sum(-1)
        return torch.clamp(x * (0.5 / math.pi**2), max=self.truncation_point)

    def log_prob(self, value):
        value = value.unsqueeze(-1)
        two_n_plus_one = (
            2.0 * torch.arange(0, self.num_log_prob_terms, device=self.prototype.device)
            + 1.0
        )
        log_terms = (
            two_n_plus_one.log()
            - 1.5 * value.log()
            - 0.125 * two_n_plus_one.pow(2.0) / value
        )
        even_terms = log_terms[..., ::2]
        odd_terms = log_terms[..., 1::2]
        sum_even = torch.logsumexp(even_terms, dim=-1).exp()
        sum_odd = torch.logsumexp(odd_terms, dim=-1).exp()
        return (sum_even - sum_odd).log() - 0.5 * math.log(2.0 * math.pi)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import copy
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Callable, Dict, Hashable, Union

import torch

from pyro.ops.welford import WelfordCovariance


class StreamingStats(ABC):
    """
    Abstract base class for streamable statistics of trees of tensors.

    Derived classes must implelement :meth:`update`, :meth:`merge`, and
    :meth:`get`.
    """

    @abstractmethod
    def update(self, sample) -> None:
        """
        Update state from a single sample.

        This mutates ``self`` and returns nothing. Updates should be
        independent of order, i.e. samples should be exchangeable.

        :param sample: A sample value which is a nested dictionary of
            :class:`torch.Tensor` leaves. This can have arbitrary nesting and
            shape shape, but assumes shape is constant across calls to
            ``.update()``.
        """
        raise NotImplementedError

    @abstractmethod
    def merge(self, other) -> "StreamingStats":
        """
        Select two aggregate statistics, e.g. from different MCMC chains.

        This is a pure function: it returns a new :class:`StreamingStats`
        object and does not modify either ``self`` or ``other``.

        :param other: Another streaming stats instance of the same type.
        """
        assert isinstance(other, type(self))
        raise NotImplementedError

    @abstractmethod
    def get(self) -> Any:
        """
        Return the aggregate statistic.
        """
        raise NotImplementedError


class CountStats(StreamingStats):
    """
    Statistic tracking only the number of samples.

    For example::

        >>> stats = CountStats()
        >>> stats.update(torch.randn(3, 3))
        >>> stats.get()
        {'count': 1}
    """

    def __init__(self):
        self.count = 0
        super().__init__()

    def update(self, sample) -> None:
        self.count += 1

    def merge(self, other: "CountStats") -> "CountStats":
        assert isinstance(other, type(self))
        result = CountStats()
        result.count = self.count + other.count
        return result

    def get(self) -> Dict[str, int]:
        """
        :returns: A dictionary with keys ``count: int``.
        :rtype: dict
        """
        return {"count": self.count}


class StatsOfDict(StreamingStats):
    """
    Statistics of samples that are dictionaries with constant set of keys.

    For example the following are equivalent::

        # Version 1. Hand encode statistics.
        >>> a_stats = CountStats()
        >>> b_stats = CountMeanStats()
        >>> a_stats.update(torch.tensor(0.))
        >>> b_stats.update(torch.tensor([1., 2.]))
        >>> summary = {"a": a_stats.get(), "b": b_stats.get()}

        # Version 2. Collect samples into dictionaries.
        >>> stats = StatsOfDict({"a": CountStats, "b": CountMeanStats})
        >>> stats.update({"a": torch.tensor(0.), "b": torch.tensor([1., 2.])})
        >>> summary = stats.get()
        >>> summary
        {'a': {'count': 1}, 'b': {'count': 1, 'mean': tensor([1., 2.])}}

    :param default: Default type of statistics of values of the dictionary.
        Defaults to the inexpensive :class:`CountStats`.
    :param dict types: Dictionary mapping key to type of statistic that should
        be recorded for values corresponding to that key.
    """

    def __init__(
        self,
        types: Dict[Hashable, Callable[[], StreamingStats]] = {},
        default: Callable[[], StreamingStats] = CountStats,
    ):
        self.stats: Dict[Hashable, StreamingStats] = defaultdict(default)
        self.stats.update({k: v() for k, v in types.items()})
        super().__init__()

    def update(self, sample: Dict[Hashable, Any]) -> None:
        for k, v in sample.items():
            self.stats[k].update(v)

    def merge(self, other: "StatsOfDict") -> "StatsOfDict":
        assert isinstance(other, type(self))
        result = copy.deepcopy(self)
        for k in set(self.stats).union(other.stats):
            if k not in self.stats:
                result.stats[k] = copy.deepcopy(other.stats[k])
            elif k in other.stats:
                result.stats[k] = self.stats[k].merge(other.stats[k])
        return result

    def get(self) -> Dict[Hashable, Any]:
        """
        :returns: A dictionary of statistics. The keys of this dictionary are
            the same as the keys of the samples from which this object is
            updated.
        :rtype: dict
        """
        return {k: v.get() for k, v in self.stats.items()}


class StackStats(StreamingStats):
    """
    Statistic collecting a stream of tensors into a single stacked tensor.
    """

    def __init__(self):
        self.samples = []

    def update(self, sample: torch.Tensor) -> None:
        assert isinstance(sample, torch.Tensor)
        self.samples.append(sample)

    def merge(self, other: "StackStats") -> "StackStats":
        assert isinstance(other, type(self))
        result = StackStats()
        result.samples = self.samples + other.samples
        return result

    def get(self) -> Dict[str, Union[int, torch.Tensor]]:
        """
        :returns: A dictionary with keys ``count: int`` and (if any samples
            have been collected) ``samples: torch.Tensor``.
        :rtype: dict
        """
        if not self.samples:
            return {"count": 0}
        return {"count": len(self.samples), "samples": torch.stack(self.samples)}


class CountMeanStats(StreamingStats):
    """
    Statistic tracking the count and mean of a single :class:`torch.Tensor`.
    """

    def __init__(self):
        self.count = 0
        self.mean = 0
        super().__init__()

    def update(self, sample: torch.Tensor) -> None:
        assert isinstance(sample, torch.Tensor)
        self.count += 1
        self.mean += (sample.detach() - self.mean) / self.count

    def merge(self, other: "CountMeanStats") -> "CountMeanStats":
        assert isinstance(other, type(self))
        result = CountMeanStats()
        result.count = self.count + other.count
        p = self.count / max(result.count, 1)
        q = other.count / max(result.count, 1)
        result.mean = p * self.mean + q * other.mean
        return result

    def get(self) -> Dict[str, Union[int, torch.Tensor]]:
        """
        :returns: A dictionary with keys ``count: int`` and (if any samples
            have been collected) ``mean: torch.Tensor``.
        :rtype: dict
        """
        if self.count == 0:
            return {"count": 0}
        return {"count": self.count, "mean": self.mean}


class CountMeanVarianceStats(StreamingStats):
    """
    Statistic tracking the count, mean, and (diagonal) variance of a single
    :class:`torch.Tensor`.
    """

    def __init__(self):
        self.shape = None
        self.welford = WelfordCovariance(diagonal=True)
        super().__init__()

    def update(self, sample: torch.Tensor) -> None:
        assert isinstance(sample, torch.Tensor)
        if self.shape is None:
            self.shape = sample.shape
        assert sample.shape == self.shape
        self.welford.update(sample.detach().reshape(-1))

    def merge(self, other: "CountMeanVarianceStats") -> "CountMeanVarianceStats":
        assert isinstance(other, type(self))
        if self.shape is None:
            return copy.deepcopy(other)
        if other.shape is None:
            return copy.deepcopy(self)
        result = copy.deepcopy(self)
        res = result.welford
        lhs = self.welford
        rhs = other.welford
        res.n_samples = lhs.n_samples + rhs.n_samples
        lhs_weight = lhs.n_samples / res.n_samples
        rhs_weight = rhs.n_samples / res.n_samples
        res._mean = lhs_weight * lhs._mean + rhs_weight * rhs._mean
        res._m2 = (
            lhs._m2
            + rhs._m2
            + (lhs.n_samples * rhs.n_samples / res.n_samples)
            * (lhs._mean - rhs._mean) ** 2
        )
        return result

    def get(self) -> Dict[str, Union[int, torch.Tensor]]:
        """
        :returns: A dictionary with keys ``count: int`` and (if any samples
            have been collected) ``mean: torch.Tensor`` and ``variance:
            torch.Tensor``.
        :rtype: dict
        """
        if self.shape is None:
            return {"count": 0}
        count = self.welford.n_samples
        mean = self.welford._mean.reshape(self.shape)
        variance = self.welford.get_covariance(regularize=False).reshape(self.shape)
        return {"count": count, "mean": mean, "variance": variance}


# Note this is ordered logically for sphinx rather than alphabetically.
__all__ = [
    "StreamingStats",
    "StatsOfDict",
    "StackStats",
    "CountStats",
    "CountMeanStats",
    "CountMeanVarianceStats",
]



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch.nn.functional as F

import pyro
import pyro.distributions as dist
from pyro.contrib.gp.likelihoods.likelihood import Likelihood


def _softmax(x):
    return F.softmax(x, dim=-1)


class MultiClass(Likelihood):
    """
    Implementation of MultiClass likelihood, which is used for multi-class
    classification problems.

    MultiClass likelihood uses :class:`~pyro.distributions.Categorical`
    distribution, so ``response_function`` should normalize its input's rightmost axis.
    By default, we use `softmax` function.

    :param int num_classes: Number of classes for prediction.
    :param callable response_function: A mapping to correct domain for MultiClass
        likelihood.
    """

    def __init__(self, num_classes, response_function=None):
        super().__init__()
        self.num_classes = num_classes
        self.response_function = (
            _softmax if response_function is None else response_function
        )

    def forward(self, f_loc, f_var, y=None):
        r"""
        Samples :math:`y` given :math:`f_{loc}`, :math:`f_{var}` according to

            .. math:: f & \sim \mathbb{Normal}(f_{loc}, f_{var}),\\
                y & \sim \mathbb{Categorical}(f).

        .. note:: The log likelihood is estimated using Monte Carlo with 1 sample of
            :math:`f`.

        :param torch.Tensor f_loc: Mean of latent function output.
        :param torch.Tensor f_var: Variance of latent function output.
        :param torch.Tensor y: Training output tensor.
        :returns: a tensor sampled from likelihood
        :rtype: torch.Tensor
        """
        # calculates Monte Carlo estimate for E_q(f) [logp(y | f)]
        f = dist.Normal(f_loc, f_var.sqrt())()
        if f.dim() < 2:
            raise ValueError(
                "Latent function output should have at least 2 "
                "dimensions: one for number of classes and one for "
                "number of data."
            )

        # swap class dimension and data dimension
        f_swap = f.transpose(-2, -1)  # -> num_data x num_classes
        if f_swap.size(-1) != self.num_classes:
            raise ValueError(
                "Number of Gaussian processes should be equal to the "
                "number of classes. Expected {} but got {}.".format(
                    self.num_classes, f_swap.size(-1)
                )
            )
        if self.response_function is _softmax:
            y_dist = dist.Categorical(logits=f_swap)
        else:
            f_res = self.response_function(f_swap)
            y_dist = dist.Categorical(f_res)
        if y is not None:
            y_dist = y_dist.expand_by(y.shape[: -f.dim() + 1]).to_event(y.dim())
        return pyro.sample(self._pyro_get_fullname("y"), y_dist, obs=y)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
# Copyright (c) 2015-2016, Gamelan Labs, Inc.
# Copyright (c) 2016, Google, Inc.
# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0 AND BSD-3-Clause

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module computes the chi^2 survival function and the required
functions.
"""

import math
import sys


def log(x):
    if x > sys.float_info.min:
        value = math.log(x)
    else:
        value = -math.inf
    return value


def incomplete_gamma(x, s):
    r"""
    This function computes the incomplete lower gamma function
    using the series expansion:

    .. math::

       \gamma(x, s) = x^s \Gamma(s)e^{-x}\sum^\infty_{k=0}
                    \frac{x^k}{\Gamma(s + k + 1)}

    This series will converge strongly because the Gamma
    function grows factorially.

    Because the Gamma function does grow so quickly, we can
    run into numerical stability issues. To solve this we carry
    out as much math as possible in the log domain to reduce
    numerical error. This function matches the results from
    scipy to numerical precision.
    """
    if x < 0:
        return 1
    if x > 1e3:
        return math.gamma(s)
    log_gamma_s = math.lgamma(s)
    log_x = log(x)
    value = 0
    for k in range(100):
        log_num = (k + s) * log_x + (-x) + log_gamma_s
        log_denom = math.lgamma(k + s + 1)
        value += math.exp(log_num - log_denom)
    return value


def chi2sf(x, s):
    r"""
    This function returns the survival function of the chi^2
    distribution. The survival function is given as:

    .. math::
       1 - CDF

    where rhe chi^2 CDF is given as

    .. math::
       F(x; s) = \frac{ \gamma( x/2, s/2 ) }{ \Gamma(s/2) },

    with :math:`\gamma` is the incomplete gamma function defined above.
    Therefore, the survival probability is givne by:

    .. math::
       1 - \frac{ \gamma( x/2, s/2 ) }{ \Gamma(s/2) }.

    This function matches the results from
    scipy to numerical precision.
    """
    top = incomplete_gamma(x / 2, s / 2)
    bottom = math.gamma(s / 2)
    value = top / bottom
    return 1 - value



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import opt_einsum

from pyro.util import ignore_jit_warnings

_PATH_CACHE = {}


def contract_expression(equation, *shapes, **kwargs):
    """
    Wrapper around :func:`opt_einsum.contract_expression` that optionally uses
    Pyro's cheap optimizer and optionally caches contraction paths.

    :param bool cache_path: whether to cache the contraction path.
        Defaults to True.
    """
    # memoize the contraction path
    cache_path = kwargs.pop("cache_path", True)
    if cache_path:
        kwargs_key = tuple(kwargs.items())
        key = equation, shapes, kwargs_key
        if key in _PATH_CACHE:
            return _PATH_CACHE[key]

    expr = opt_einsum.contract_expression(equation, *shapes, **kwargs)
    if cache_path:
        _PATH_CACHE[key] = expr
    return expr


def contract(equation, *operands, **kwargs):
    """
    Wrapper around :func:`opt_einsum.contract` that optionally uses Pyro's
    cheap optimizer and optionally caches contraction paths.

    :param bool cache_path: whether to cache the contraction path.
        Defaults to True.
    """
    backend = kwargs.pop("backend", "numpy")
    out = kwargs.pop("out", None)
    shapes = [tuple(t.shape) for t in operands]
    with ignore_jit_warnings():
        expr = contract_expression(equation, *shapes)
        return expr(*operands, backend=backend, out=out)


__all__ = ["contract", "contract_expression"]



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from functools import partial
from typing import TYPE_CHECKING, Callable, List, Optional

from pyro.poutine.messenger import Messenger

if TYPE_CHECKING:
    from pyro.poutine.runtime import Message


def _block_fn(
    expose: List[str],
    expose_types: List[str],
    hide: List[str],
    hide_types: List[str],
    hide_all: bool,
    msg: "Message",
) -> bool:
    # handle observes
    if msg["type"] == "sample" and msg["is_observed"]:
        msg_type = "observe"
    else:
        msg_type = msg["type"]

    is_not_exposed = (msg["name"] not in expose) and (msg_type not in expose_types)

    # decision rule for hiding:
    if (
        (msg["name"] in hide)
        or (msg_type in hide_types)
        or (is_not_exposed and hide_all)
    ):  # noqa: E129
        return True
    # otherwise expose
    else:
        return False


def _make_default_hide_fn(
    hide_all: bool,
    expose_all: bool,
    hide: Optional[List[str]],
    expose: Optional[List[str]],
    hide_types: Optional[List[str]],
    expose_types: Optional[List[str]],
) -> Callable[["Message"], bool]:
    # first, some sanity checks:
    # hide_all and expose_all intersect?
    assert (hide_all is False and expose_all is False) or (
        hide_all != expose_all
    ), "cannot hide and expose a site"

    # hide and expose intersect?
    if hide is None:
        hide = []
    else:
        hide_all = False

    if expose is None:
        expose = []
    else:
        hide_all = True

    assert set(hide).isdisjoint(set(expose)), "cannot hide and expose a site"

    # hide_types and expose_types intersect?
    if hide_types is None:
        hide_types = []
    else:
        hide_all = False

    if expose_types is None:
        expose_types = []
    else:
        hide_all = True

    assert set(hide_types).isdisjoint(
        set(expose_types)
    ), "cannot hide and expose a site type"

    return partial(_block_fn, expose, expose_types, hide, hide_types, hide_all)


def _negate_fn(
    fn: Callable[["Message"], Optional[bool]]
) -> Callable[["Message"], bool]:
    # typed version of lambda msg: not fn(msg)
    def negated_fn(msg: "Message") -> bool:
        return not fn(msg)

    return negated_fn


class BlockMessenger(Messenger):
    """
    This handler selectively hides Pyro primitive sites from the outside world.
    Default behavior: block everything.

    A site is hidden if at least one of the following holds:

        0. ``hide_fn(msg) is True`` or ``(not expose_fn(msg)) is True``
        1. ``msg["name"] in hide``
        2. ``msg["type"] in hide_types``
        3. ``msg["name"] not in expose and msg["type"] not in expose_types``
        4. ``hide``, ``hide_types``, and ``expose_types`` are all ``None``

    For example, suppose the stochastic function fn has two sample sites "a" and "b".
    Then any effect outside of ``BlockMessenger(fn, hide=["a"])``
    will not be applied to site "a" and will only see site "b":

        >>> def fn():
        ...     a = pyro.sample("a", dist.Normal(0., 1.))
        ...     return pyro.sample("b", dist.Normal(a, 1.))
        >>> fn_inner = pyro.poutine.trace(fn)
        >>> fn_outer = pyro.poutine.trace(pyro.poutine.block(fn_inner, hide=["a"]))
        >>> trace_inner = fn_inner.get_trace()
        >>> trace_outer  = fn_outer.get_trace()
        >>> "a" in trace_inner
        True
        >>> "a" in trace_outer
        False
        >>> "b" in trace_inner
        True
        >>> "b" in trace_outer
        True

    :param fn: a stochastic function (callable containing Pyro primitive calls)
    :param hide_fn: function that takes a site and returns True to hide the site
        or False/None to expose it.  If specified, all other parameters are ignored.
        Only specify one of hide_fn or expose_fn, not both.
    :param expose_fn: function that takes a site and returns True to expose the site
        or False/None to hide it.  If specified, all other parameters are ignored.
        Only specify one of hide_fn or expose_fn, not both.
    :param bool hide_all: hide all sites
    :param bool expose_all: expose all sites normally
    :param list hide: list of site names to hide
    :param list expose: list of site names to be exposed while all others hidden
    :param list hide_types: list of site types to be hidden
    :param list expose_types: list of site types to be exposed while all others hidden
    :returns: stochastic function decorated with a :class:`~pyro.poutine.block_messenger.BlockMessenger`
    """

    def __init__(
        self,
        hide_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
        expose_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
        hide_all: bool = True,
        expose_all: bool = False,
        hide: Optional[List[str]] = None,
        expose: Optional[List[str]] = None,
        hide_types: Optional[List[str]] = None,
        expose_types: Optional[List[str]] = None,
    ) -> None:
        super().__init__()
        if not (hide_fn is None or expose_fn is None):
            raise ValueError("Only specify one of hide_fn or expose_fn")
        if hide_fn is not None:
            self.hide_fn = hide_fn
        elif expose_fn is not None:
            self.hide_fn = _negate_fn(expose_fn)
        else:
            self.hide_fn = _make_default_hide_fn(
                hide_all, expose_all, hide, expose, hide_types, expose_types
            )

    def _process_message(self, msg: "Message") -> None:
        msg["stop"] = bool(self.hide_fn(msg))



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

r"""
Contributed Code
================

.. warning:: Code in ``pyro.contrib`` is under various stages of development.
    This code makes no guarantee about maintaining backwards compatibility.
"""

from pyro.contrib import (
    autoname,
    bnn,
    easyguide,
    epidemiology,
    forecast,
    gp,
    oed,
    tracking,
)

__all__ = [
    "autoname",
    "bnn",
    "easyguide",
    "epidemiology",
    "forecast",
    "gp",
    "oed",
    "tracking",
]


try:
    import funsor as funsor_  # noqa: F401

    from pyro.contrib import funsor

    __all__ += ["funsor"]
except ImportError:
    pass



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import constraints
from ..torch_transform import TransformModule
from ..util import copy_docs_from


@copy_docs_from(TransformModule)
class BatchNorm(TransformModule):
    r"""
    A type of batch normalization that can be used to stabilize training in
    normalizing flows. The inverse operation is defined as

        :math:`x = (y - \hat{\mu}) \oslash \sqrt{\hat{\sigma^2}} \otimes \gamma + \beta`

    that is, the standard batch norm equation, where :math:`x` is the input,
    :math:`y` is the output, :math:`\gamma,\beta` are learnable parameters, and
    :math:`\hat{\mu}`/:math:`\hat{\sigma^2}` are smoothed running averages of
    the sample mean and variance, respectively. The constraint :math:`\gamma>0` is
    enforced to ease calculation of the log-det-Jacobian term.

    This is an element-wise transform, and when applied to a vector, learns two
    parameters (:math:`\gamma,\beta`) for each dimension of the input.

    When the module is set to training mode, the moving averages of the sample mean
    and variance are updated every time the inverse operator is called, e.g., when a
    normalizing flow scores a minibatch with the `log_prob` method.

    Also, when the module is set to training mode, the sample mean and variance on
    the current minibatch are used in place of the smoothed averages,
    :math:`\hat{\mu}` and :math:`\hat{\sigma^2}`, for the inverse operator. For
    this reason it is not the case that :math:`x=g(g^{-1}(x))` during training,
    i.e., that the inverse operation is the inverse of the forward one.

    Example usage:

    >>> from pyro.nn import AutoRegressiveNN
    >>> from pyro.distributions.transforms import AffineAutoregressive
    >>> base_dist = dist.Normal(torch.zeros(10), torch.ones(10))
    >>> iafs = [AffineAutoregressive(AutoRegressiveNN(10, [40])) for _ in range(2)]
    >>> bn = BatchNorm(10)
    >>> flow_dist = dist.TransformedDistribution(base_dist, [iafs[0], bn, iafs[1]])
    >>> flow_dist.sample()  # doctest: +SKIP

    :param input_dim: the dimension of the input
    :type input_dim: int
    :param momentum: momentum parameter for updating moving averages
    :type momentum: float
    :param epsilon: small number to add to variances to ensure numerical stability
    :type epsilon: float

    References:

    [1] Sergey Ioffe and Christian Szegedy. Batch Normalization: Accelerating Deep
    Network Training by Reducing Internal Covariate Shift. In International
    Conference on Machine Learning, 2015. https://arxiv.org/abs/1502.03167

    [2] Laurent Dinh, Jascha Sohl-Dickstein, and Samy Bengio. Density Estimation
    using Real NVP. In International Conference on Learning Representations, 2017.
    https://arxiv.org/abs/1605.08803

    [3] George Papamakarios, Theo Pavlakou, and Iain Murray. Masked Autoregressive
    Flow for Density Estimation. In Neural Information Processing Systems, 2017.
    https://arxiv.org/abs/1705.07057

    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True

    def __init__(self, input_dim, momentum=0.1, epsilon=1e-5):
        super().__init__()

        self.input_dim = input_dim
        self.gamma = nn.Parameter(torch.ones(input_dim))
        self.beta = nn.Parameter(torch.zeros(input_dim))
        self.momentum = momentum
        self.epsilon = epsilon

        self.register_buffer("moving_mean", torch.zeros(input_dim))
        self.register_buffer("moving_variance", torch.ones(input_dim))

    @property
    def constrained_gamma(self):
        return F.relu(self.gamma) + 1e-6

    def _call(self, x):
        """
        :param x: the input into the bijection
        :type x: torch.Tensor

        Invokes the bijection x=>y; in the prototypical context of a
        :class:`~pyro.distributions.TransformedDistribution` `x` is a sample from
        the base distribution (or the output of a previous transform)
        """
        # Enforcing the constraint that gamma is positive
        return (x - self.beta) / self.constrained_gamma * torch.sqrt(
            self.moving_variance + self.epsilon
        ) + self.moving_mean

    def _inverse(self, y):
        """
        :param y: the output of the bijection
        :type y: torch.Tensor

        Inverts y => x.
        """
        # During training, keep smoothed average of sample mean and variance
        if self.training:
            mean, var = y.mean(0), y.var(0)

            with torch.no_grad():
                # NOTE: The momentum variable agrees with the definition in e.g. `torch.nn.BatchNorm1d`
                self.moving_mean.mul_(1 - self.momentum).add_(mean * self.momentum)
                self.moving_variance.mul_(1 - self.momentum).add_(var * self.momentum)

        # During test time, use smoothed averages rather than the sample ones
        else:
            mean, var = self.moving_mean, self.moving_variance

        return (y - mean) * self.constrained_gamma / torch.sqrt(
            var + self.epsilon
        ) + self.beta

    def log_abs_det_jacobian(self, x, y):
        """
        Calculates the elementwise determinant of the log Jacobian, dx/dy
        """
        if self.training:
            var = torch.var(y, dim=0, keepdim=True)
        else:
            # NOTE: You wouldn't typically run this function in eval mode, but included for gradient tests
            var = self.moving_variance
        return -self.constrained_gamma.log() + 0.5 * torch.log(var + self.epsilon)


def batchnorm(input_dim, **kwargs):
    """
    A helper function to create a :class:`~pyro.distributions.transforms.BatchNorm`
    object for consistency with other helpers.

    :param input_dim: Dimension of input variable
    :type input_dim: int
    :param momentum: momentum parameter for updating moving averages
    :type momentum: float
    :param epsilon: small number to add to variances to ensure numerical stability
    :type epsilon: float

    """
    bn = BatchNorm(input_dim, **kwargs)
    return bn



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings

import pytest
import torch

from pyro import util

pytestmark = pytest.mark.stage("unit")


def test_warn_if_nan():
    # scalar case
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        x = float("inf")
        msg = "example message"
        y = util.warn_if_nan(x, msg)
        assert y is x
        assert len(w) == 0

        x = float("nan")
        util.warn_if_nan(x, msg)
        # Verify some things
        assert len(w) == 1
        assert msg in str(w[-1].message)

    # tensor case
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        x = torch.ones(2)
        msg = "example message"
        util.warn_if_nan(x, msg)
        x[1] = float("nan")
        util.warn_if_nan(x, msg)
        assert len(w) == 1
        assert msg in str(w[-1].message)

    # grad case
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        x = torch.ones(2, requires_grad=True)
        util.warn_if_nan(x, msg)
        y = x.sum()
        y.backward([torch.tensor(float("nan"))])
        assert len(w) == 1
        assert msg in str(w[-1].message)


def test_warn_if_inf():
    # scalar case
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        x = 3
        msg = "example message"
        y = util.warn_if_inf(x, msg, allow_posinf=True, allow_neginf=True)
        assert y is x
        assert len(w) == 0
        x = float("inf")
        util.warn_if_inf(x, msg, allow_posinf=True)
        assert len(w) == 0
        util.warn_if_inf(x, msg, allow_neginf=True)
        assert len(w) == 1
        assert msg in str(w[-1].message)

    # tensor case
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        x = torch.ones(2)
        util.warn_if_inf(x, msg, allow_posinf=True, allow_neginf=True)
        assert len(w) == 0
        x[0] = float("inf")
        util.warn_if_inf(x, msg, allow_posinf=True)
        assert len(w) == 0
        util.warn_if_inf(x, msg, allow_neginf=True)
        assert len(w) == 1
        assert msg in str(w[-1].message)

    # grad case
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        x = torch.ones(2, requires_grad=True)
        util.warn_if_inf(x, msg, allow_posinf=True)
        y = x.sum()
        y.backward([torch.tensor(float("inf"))])
        assert len(w) == 0

        x.grad = None
        y.backward([torch.tensor(-float("inf"))])
        assert len(w) == 1
        assert msg in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        z = torch.ones(2, requires_grad=True)
        y = z.sum()
        util.warn_if_inf(z, msg, allow_neginf=True)
        y.backward([torch.tensor(-float("inf"))])
        assert len(w) == 0
        z.grad = None
        y.backward([torch.tensor(float("inf"))])
        assert len(w) == 1
        assert msg in str(w[-1].message)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import math
from typing import Callable, Dict, Optional, Tuple

import torch
from torch.nn.functional import pad
from torch.optim.optimizer import Optimizer

from pyro.ops.tensor_utils import dct, idct, next_fast_len


def _transform_forward(x: torch.Tensor, dim: int, duration: int) -> torch.Tensor:
    assert not x.requires_grad
    assert dim < 0
    assert duration == x.size(dim)
    time_domain = x
    new_size = next_fast_len(duration)
    if new_size == duration:
        freq_domain = x
    else:
        freq_domain = pad(x, (0, 0) * (-1 - dim) + (0, new_size - duration))
    freq_domain = dct(freq_domain, dim)
    return torch.cat([time_domain, freq_domain], dim=dim)


def _transform_inverse(x: torch.Tensor, dim: int, duration: int):
    assert not x.requires_grad
    assert dim < 0
    dots = (slice(None),) * (x.dim() + dim)
    left = dots + (slice(None, duration),)
    right = dots + (slice(duration, None),)
    return idct(x[right], dim)[left].add_(x[left])


def _get_mask(x, indices):
    # create a boolean mask from subsample indices:
    #   + start with a zero tensor
    #   + at a specific `dim`, we increase by 1
    #     all entries which have index at `dim`
    #     belong to `indices[dim]`
    #   + at the end, we will get a tensor whose
    #     values at `indices` are `len(indices)`
    mask = x.new_zeros(x.shape, dtype=torch.long, device=x.device)
    if len(indices) > 0:
        idx = []
        for dim in range(-x.dim(), 0):
            if dim in indices:
                mask[idx + [indices[dim]]] += 1
            idx.append(slice(None))
    return mask == len(indices)


class DCTAdam(Optimizer):
    """
    EXPERIMENTAL Discrete Cosine Transform-augmented
    :class:`~pyro.optim.clipped_adam.ClippedAdam` optimizer.

    This acts like :class:`~pyro.optim.clipped_adam.ClippedAdam` on most
    parameters, but if a parameter has an attribute ``._pyro_dct_dim``
    indicating a time dimension, this creates a secondary optimize in the
    frequency domain. This is useful for parameters of time series models.

    :param params: iterable of parameters to optimize or dicts defining parameter groups
    :param float lr: learning rate (default: 1e-3)
    :param tuple betas: coefficients used for computing
        running averages of gradient and its square (default: (0.9, 0.999))
    :param float eps: term added to the denominator to improve
        numerical stability (default: 1e-8)
    :param float clip_norm: magnitude of norm to which gradients are clipped (default: 10.0)
    :param float lrd: rate at which learning rate decays (default: 1.0)
    :param bool subsample_aware: whether to update gradient statistics only for
        those elements that appear in a subsample (default: False).
    """

    def __init__(
        self,
        params,
        lr: float = 1e-3,
        betas: Tuple = (0.9, 0.999),
        eps: float = 1e-8,
        clip_norm: float = 10.0,
        lrd: float = 1.0,
        subsample_aware: bool = False,
    ):
        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            clip_norm=clip_norm,
            lrd=lrd,
            subsample_aware=subsample_aware,
        )
        super().__init__(params, defaults)

    def step(self, closure: Optional[Callable] = None) -> Optional[float]:
        """
        :param closure: An optional closure that reevaluates the model and returns the loss.

        Performs a single optimization step.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            group["lr"] *= group["lrd"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                subsample = getattr(p, "_pyro_subsample", {})
                if subsample and group["subsample_aware"]:
                    self._step_param_subsample(group, p, subsample)
                else:
                    self._step_param(group, p)

        return loss

    def _step_param(self, group: Dict, p) -> None:
        grad = p.grad.data
        grad.clamp_(-group["clip_norm"], group["clip_norm"])

        # Transform selected parameters via dct.
        time_dim = getattr(p, "_pyro_dct_dim", None)
        if time_dim is not None:
            duration = p.size(time_dim)
            grad = _transform_forward(grad, time_dim, duration)

        state = self.state[p]

        # State initialization
        if len(state) == 0:
            state["step"] = 0
            # Exponential moving average of gradient values
            state["exp_avg"] = torch.zeros_like(grad)
            # Exponential moving average of squared gradient values
            state["exp_avg_sq"] = torch.zeros_like(grad)

        exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
        beta1, beta2 = group["betas"]

        state["step"] += 1

        # Decay the first and second moment running average coefficient
        exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
        exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

        denom = exp_avg_sq.sqrt().add_(group["eps"])

        bias_correction1 = 1 - beta1 ** state["step"]
        bias_correction2 = 1 - beta2 ** state["step"]
        step_size = group["lr"] * math.sqrt(bias_correction2) / bias_correction1

        if time_dim is None:
            p.data.addcdiv_(exp_avg, denom, value=-step_size)
        else:
            step = _transform_inverse(exp_avg / denom, time_dim, duration)
            p.data.add_(step.mul_(-step_size))

    def _step_param_subsample(self, group: Dict, p, subsample) -> None:
        mask = _get_mask(p, subsample)

        grad = p.grad.data.masked_select(mask)
        grad.clamp_(-group["clip_norm"], group["clip_norm"])

        # Transform selected parameters via dct.
        time_dim = getattr(p, "_pyro_dct_dim", None)
        if time_dim is not None:
            duration = p.size(time_dim)
            grad = _transform_forward(grad, time_dim, duration)

        state = self.state[p]

        # State initialization
        if len(state) == 0:
            state["step"] = torch.zeros_like(p)
            # Exponential moving average of gradient values
            state["exp_avg"] = torch.zeros_like(p)
            # Exponential moving average of squared gradient values
            state["exp_avg_sq"] = torch.zeros_like(p)

        beta1, beta2 = group["betas"]

        state_step = state["step"].masked_select(mask).add_(1)
        state["step"].masked_scatter_(mask, state_step)

        # Decay the first and second moment running average coefficient
        exp_avg = (
            state["exp_avg"].masked_select(mask).mul_(beta1).add_(grad, alpha=1 - beta1)
        )
        state["exp_avg"].masked_scatter_(mask, exp_avg)

        exp_avg_sq = (
            state["exp_avg_sq"]
            .masked_select(mask)
            .mul_(beta2)
            .addcmul_(grad, grad, value=1 - beta2)
        )
        state["exp_avg_sq"].masked_scatter_(mask, exp_avg_sq)

        denom = exp_avg_sq.sqrt_().add_(group["eps"])

        bias_correction1 = 1 - beta1**state_step
        bias_correction2 = 1 - beta2**state_step
        step_size = bias_correction2.sqrt_().div_(bias_correction1).mul_(group["lr"])

        step = exp_avg.div_(denom)
        if time_dim is not None:
            step = _transform_inverse(step, time_dim, duration)
        p_data = p.data.masked_select(mask).sub_(step.mul_(step_size))
        p.data.masked_scatter_(mask, p_data)
